#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import time

from oslo_concurrency import lockutils
from oslo_concurrency import processutils as putils
from oslo_log import log as logging

from os_brick import exception
from os_brick import initiator
from os_brick.initiator.connectors import base
from os_brick import utils


DEVICE_SCAN_ATTEMPTS_DEFAULT = 3

LOG = logging.getLogger(__name__)

synchronized = lockutils.synchronized_with_prefix('os-brick-')


class NVMeConnector(base.BaseLinuxConnector):

    """Connector class to attach/detach NVMe over fabric volumes."""

    def __init__(self, root_helper, driver=None,
                 device_scan_attempts=initiator.DEVICE_SCAN_ATTEMPTS_DEFAULT,
                 *args, **kwargs):
        super(NVMeConnector, self).__init__(
            root_helper,
            driver=driver,
            device_scan_attempts=device_scan_attempts,
            *args, **kwargs)

    @staticmethod
    def get_connector_properties(root_helper, *args, **kwargs):
        """The NVMe connector properties."""
        return {}

    def get_search_path(self):
        return '/dev'

    def get_volume_paths(self, connection_properties):
        path = connection_properties['device_path']
        LOG.debug("Path of volume to be extended is %(path)s", {'path': path})
        return [path]

    def _get_nvme_devices(self):
        nvme_devices = []
        pattern = r'/dev/nvme[0-9]n[0-9]'
        cmd = ['nvme', 'list']
        try:
            (out, err) = self._execute(*cmd, root_helper=self._root_helper,
                                       run_as_root=True)
            for line in out.split('\n'):
                result = re.match(pattern, line)
                if result:
                    nvme_devices.append(result.group(0))
            LOG.debug("_get_nvme_devices returned %(nvme_devices)s",
                      {'nvme_devices': nvme_devices})
            return nvme_devices

        except putils.ProcessExecutionError:
            LOG.error(
                "Failed to list available NVMe connected controllers.")
            raise

    @utils.trace
    @synchronized('connect_volume')
    def connect_volume(self, connection_properties):
        """Discover and attach the volume.

        :param connection_properties: The dictionary that describes all
                                      of the target volume attributes.
               connection_properties must include:
               nqn - NVMe subsystem name to the volume to be connected
               target_port - NVMe target port that hosts the nqn sybsystem
               target_portal - NVMe target ip that hosts the nqn sybsystem
        :type connection_properties: dict
        :returns: dict
        """

        current_nvme_devices = self._get_nvme_devices()

        device_info = {'type': 'block'}
        conn_nqn = connection_properties['nqn']
        target_portal = connection_properties['target_portal']
        port = connection_properties['target_port']
        nvme_transport_type = connection_properties['transport_type']
        cmd = [
            'nvme', 'connect',
            '-t', nvme_transport_type,
            '-n', conn_nqn,
            '-a', target_portal,
            '-s', port]
        try:
            self._execute(*cmd, root_helper=self._root_helper,
                          run_as_root=True)
        except putils.ProcessExecutionError:
            LOG.error(
                "Failed to connect to NVMe nqn "
                "%(conn_nqn)s", {'conn_nqn': conn_nqn})
            raise

        for retry in range(1, self.device_scan_attempts + 1):
            all_nvme_devices = self._get_nvme_devices()
            path = set(all_nvme_devices) - set(current_nvme_devices)
            if path:
                break
            time.sleep(retry ** 2)
        else:
            LOG.error("Failed to retrieve available connected NVMe "
                      "controllers when running nvme list")
            raise exception.TargetPortalNotFound(target_portal)

        path = list(path)
        LOG.debug("all_nvme_devices are %(all_nvme_devices)s",
                  {'all_nvme_devices': all_nvme_devices})
        device_info['path'] = path[0]
        LOG.debug("NVMe device to be connected to is %(path)s",
                  {'path': path[0]})
        return device_info

    @utils.trace
    @synchronized('disconnect_volume')
    def disconnect_volume(self, connection_properties, device_info):
        """Detach and flush the volume.

        :param connection_properties: The dictionary that describes all
                                      of the target volume attributes.
               connection_properties must include:
               device_path - path to the volume to be connected
        :type connection_properties: dict

        :param device_info: historical difference, but same as connection_props
        :type device_info: dict

        """

        conn_nqn = connection_properties['nqn']
        device_path = connection_properties['device_path']
        LOG.debug(
            "Trying to disconnect from NVMe nqn "
            "%(conn_nqn)s with device_path %(device_path)s",
            {'conn_nqn': conn_nqn, 'device_path': device_path})
        cmd = [
            'nvme',
            'disconnect',
            '-d',
            device_path]
        try:
            self._execute(
                *cmd,
                root_helper=self._root_helper,
                run_as_root=True)

        except putils.ProcessExecutionError:
            LOG.error(
                "Failed to disconnect from NVMe nqn "
                "%(conn_nqn)s with device_path %(device_path)s",
                {'conn_nqn': conn_nqn, 'device_path': device_path})
            raise

    @utils.trace
    @synchronized('extend_volume')
    def extend_volume(self, connection_properties):
        """Update the local kernel's size information.

        Try and update the local kernel's size information
        for an LVM volume.
        """
        volume_paths = self.get_volume_paths(connection_properties)
        if volume_paths:
            return self._linuxscsi.extend_volume(volume_paths)
        else:
            LOG.warning("Couldn't find any volume paths on the host to "
                        "extend volume for %(props)s",
                        {'props': connection_properties})
            raise exception.VolumePathsNotFound()



# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
import mock
import os

import ddt
from oslo_concurrency import processutils as putils

from os_brick import exception
from os_brick.initiator.connectors import iscsi
from os_brick.initiator import linuxscsi
from os_brick.privileged import rootwrap as priv_rootwrap
from os_brick.tests.initiator import test_connector


@ddt.ddt
class ISCSIConnectorTestCase(test_connector.ConnectorTestCase):
    SINGLE_CON_PROPS = {'volume_id': 'vol_id',
                        'target_portal': 'ip1:port1',
                        'target_iqn': 'tgt1',
                        'encryption': False,
                        'target_lun': '1'}
    CON_PROPS = {
        'volume_id': 'vol_id',
        'target_portal': 'ip1:port1',
        'target_iqn': 'tgt1',
        'target_lun': 4,
        'target_portals': ['ip1:port1', 'ip2:port2', 'ip3:port3',
                           'ip4:port4'],
        'target_iqns': ['tgt1', 'tgt2', 'tgt3', 'tgt4'],
        'target_luns': [4, 5, 6, 7],
    }

    def setUp(self):
        super(ISCSIConnectorTestCase, self).setUp()
        self.connector = iscsi.ISCSIConnector(
            None, execute=self.fake_execute, use_multipath=False)

        self.connector_with_multipath = iscsi.ISCSIConnector(
            None, execute=self.fake_execute, use_multipath=True)
        self.mock_object(self.connector._linuxscsi, 'get_name_from_path',
                         return_value="/dev/sdb")
        self._fake_iqn = 'iqn.1234-56.foo.bar:01:23456789abc'
        self._name = 'volume-00000001'
        self._iqn = 'iqn.2010-10.org.openstack:%s' % self._name
        self._location = '10.0.2.15:3260'
        self._lun = 1

    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsi_session')
    def test_get_iscsi_sessions_full(self, sessions_mock):
        iscsiadm_result = ('tcp: [session1] ip1:port1,1 tgt1 (non-flash)\n'
                           'tcp: [session2] ip2:port2,-1 tgt2 (non-flash)\n'
                           'tcp: [session3] ip3:port3,1 tgt3\n')
        sessions_mock.return_value = (iscsiadm_result, '')
        res = self.connector._get_iscsi_sessions_full()
        expected = [('tcp:', 'session1', 'ip1:port1', '1', 'tgt1'),
                    ('tcp:', 'session2', 'ip2:port2', '-1', 'tgt2'),
                    ('tcp:', 'session3', 'ip3:port3', '1', 'tgt3')]
        self.assertListEqual(expected, res)

    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsi_session',
                       return_value=(None, 'error'))
    def test_get_iscsi_sessions_full_error(self, sessions_mock):
        res = self.connector._get_iscsi_sessions_full()
        self.assertEqual([], res)
        sessions_mock.assert_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full')
    def test_get_iscsi_sessions(self, sessions_mock):
        sessions_mock.return_value = [
            ('tcp:', 'session1', 'ip1:port1', '1', 'tgt1'),
            ('tcp:', 'session2', 'ip2:port2', '-1', 'tgt2'),
            ('tcp:', 'session3', 'ip3:port3', '1', 'tgt3')]
        res = self.connector._get_iscsi_sessions()
        expected = ['ip1:port1', 'ip2:port2', 'ip3:port3']
        self.assertListEqual(expected, res)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full',
                       return_value=[])
    def test_get_iscsi_sessions_no_sessions(self, sessions_mock):
        res = self.connector._get_iscsi_sessions()
        self.assertListEqual([], res)
        sessions_mock.assert_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_execute')
    def test_get_iscsi_nodes(self, exec_mock):
        iscsiadm_result = ('ip1:port1,1 tgt1\nip2:port2,-1 tgt2\n'
                           'ip3:port3,1 tgt3\n')
        exec_mock.return_value = (iscsiadm_result, '')
        res = self.connector._get_iscsi_nodes()
        expected = [('ip1:port1', 'tgt1'), ('ip2:port2', 'tgt2'),
                    ('ip3:port3', 'tgt3')]
        self.assertListEqual(expected, res)
        exec_mock.assert_called_once_with(
            'iscsiadm', '-m', 'node', run_as_root=True,
            root_helper=self.connector._root_helper, check_exit_code=False)

    @mock.patch.object(iscsi.ISCSIConnector, '_execute')
    def test_get_iscsi_nodes_error(self, exec_mock):
        exec_mock.return_value = (None, 'error')
        res = self.connector._get_iscsi_nodes()
        self.assertEqual([], res)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_ips_iqns_luns')
    @mock.patch('glob.glob')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_nodes')
    def test_get_connection_devices(self, nodes_mock, sessions_mock,
                                    glob_mock, iql_mock):
        self.connector.use_multipath = True
        iql_mock.return_value = self.connector._get_all_targets(self.CON_PROPS)

        # List sessions from other targets and non tcp sessions
        sessions_mock.return_value = [
            ('non-tcp:', '0', 'ip1:port1', '1', 'tgt1'),
            ('tcp:', '1', 'ip1:port1', '1', 'tgt1'),
            ('tcp:', '2', 'ip2:port2', '-1', 'tgt2'),
            ('tcp:', '3', 'ip1:port1', '1', 'tgt4'),
            ('tcp:', '4', 'ip2:port2', '-1', 'tgt5')]
        # List 1 node without sessions
        nodes_mock.return_value = [('ip1:port1', 'tgt1'),
                                   ('ip2:port2', 'tgt2'),
                                   ('ip3:port3', 'tgt3')]
        sys_cls = '/sys/class/scsi_host/host'
        glob_mock.side_effect = [
            [sys_cls + '1/device/session1/target6/1:2:6:4/block/sda',
             sys_cls + '1/device/session1/target6/1:2:6:4/block/sda1'],
            [sys_cls + '2/device/session2/target7/2:2:7:5/block/sdb',
             sys_cls + '2/device/session2/target7/2:2:7:4/block/sdc'],
        ]
        res = self.connector._get_connection_devices(self.CON_PROPS)
        expected = {('ip1:port1', 'tgt1'): ({'sda'}, set()),
                    ('ip2:port2', 'tgt2'): ({'sdb'}, {'sdc'}),
                    ('ip3:port3', 'tgt3'): (set(), set())}
        self.assertDictEqual(expected, res)
        iql_mock.assert_called_once_with(self.CON_PROPS, discover=False)

    @mock.patch('glob.glob')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_nodes')
    def test_get_connection_devices_with_iqns(self, nodes_mock, sessions_mock,
                                              glob_mock):
        ips_iqns_luns = self.connector._get_all_targets(self.CON_PROPS)

        # List sessions from other targets and non tcp sessions
        sessions_mock.return_value = [
            ('non-tcp:', '0', 'ip1:port1', '1', 'tgt1'),
            ('tcp:', '1', 'ip1:port1', '1', 'tgt1'),
            ('tcp:', '2', 'ip2:port2', '-1', 'tgt2'),
            ('tcp:', '3', 'ip1:port1', '1', 'tgt4'),
            ('tcp:', '4', 'ip2:port2', '-1', 'tgt5')]
        # List 1 node without sessions
        nodes_mock.return_value = [('ip1:port1', 'tgt1'),
                                   ('ip2:port2', 'tgt2'),
                                   ('ip3:port3', 'tgt3')]
        sys_cls = '/sys/class/scsi_host/host'
        glob_mock.side_effect = [
            [sys_cls + '1/device/session1/target6/1:2:6:4/block/sda',
             sys_cls + '1/device/session1/target6/1:2:6:4/block/sda1'],
            [sys_cls + '2/device/session2/target7/2:2:7:5/block/sdb',
             sys_cls + '2/device/session2/target7/2:2:7:4/block/sdc'],
        ]
        with mock.patch.object(iscsi.ISCSIConnector,
                               '_get_all_targets') as get_targets_mock:
            res = self.connector._get_connection_devices(mock.sentinel.props,
                                                         ips_iqns_luns)
        expected = {('ip1:port1', 'tgt1'): ({'sda'}, set()),
                    ('ip2:port2', 'tgt2'): ({'sdb'}, {'sdc'}),
                    ('ip3:port3', 'tgt3'): (set(), set())}
        self.assertDictEqual(expected, res)
        get_targets_mock.assert_not_called()

    def generate_device(self, location, iqn, transport=None, lun=1):
        dev_format = "ip-%s-iscsi-%s-lun-%s" % (location, iqn, lun)
        if transport:
            dev_format = "pci-0000:00:00.0-" + dev_format
        fake_dev_path = "/dev/disk/by-path/" + dev_format
        return fake_dev_path

    def iscsi_connection(self, volume, location, iqn):
        return {
            'driver_volume_type': 'iscsi',
            'data': {
                'volume_id': volume['id'],
                'target_portal': location,
                'target_iqn': iqn,
                'target_lun': 1,
            }
        }

    def iscsi_connection_multipath(self, volume, locations, iqns, luns):
        return {
            'driver_volume_type': 'iscsi',
            'data': {
                'volume_id': volume['id'],
                'target_portals': locations,
                'target_iqns': iqns,
                'target_luns': luns,
            }
        }

    def iscsi_connection_chap(self, volume, location, iqn, auth_method,
                              auth_username, auth_password,
                              discovery_auth_method, discovery_auth_username,
                              discovery_auth_password):
        return {
            'driver_volume_type': 'iscsi',
            'data': {
                'auth_method': auth_method,
                'auth_username': auth_username,
                'auth_password': auth_password,
                'discovery_auth_method': discovery_auth_method,
                'discovery_auth_username': discovery_auth_username,
                'discovery_auth_password': discovery_auth_password,
                'target_lun': 1,
                'volume_id': volume['id'],
                'target_iqn': iqn,
                'target_portal': location,
            }
        }

    def _initiator_get_text(self, *arg, **kwargs):
        text = ('## DO NOT EDIT OR REMOVE THIS FILE!\n'
                '## If you remove this file, the iSCSI daemon '
                'will not start.\n'
                '## If you change the InitiatorName, existing '
                'access control lists\n'
                '## may reject this initiator.  The InitiatorName must '
                'be unique\n'
                '## for each iSCSI initiator.  Do NOT duplicate iSCSI '
                'InitiatorNames.\n'
                'InitiatorName=%s' % self._fake_iqn)
        return text, None

    def test_get_initiator(self):
        def initiator_no_file(*args, **kwargs):
            raise putils.ProcessExecutionError('No file')

        self.connector._execute = initiator_no_file
        initiator = self.connector.get_initiator()
        self.assertIsNone(initiator)
        self.connector._execute = self._initiator_get_text
        initiator = self.connector.get_initiator()
        self.assertEqual(initiator, self._fake_iqn)

    def test_get_connector_properties(self):
        with mock.patch.object(priv_rootwrap, 'execute') as mock_exec:
            mock_exec.return_value = self._initiator_get_text()
            multipath = True
            enforce_multipath = True
            props = iscsi.ISCSIConnector.get_connector_properties(
                'sudo', multipath=multipath,
                enforce_multipath=enforce_multipath)

            expected_props = {'initiator': self._fake_iqn}
            self.assertEqual(expected_props, props)

    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsiadm_bare')
    def test_brick_iscsi_validate_transport(self, mock_iscsiadm):
        sample_output = ('# BEGIN RECORD 2.0-872\n'
                         'iface.iscsi_ifacename = %s.fake_suffix\n'
                         'iface.net_ifacename = <empty>\n'
                         'iface.ipaddress = <empty>\n'
                         'iface.hwaddress = 00:53:00:00:53:00\n'
                         'iface.transport_name = %s\n'
                         'iface.initiatorname = <empty>\n'
                         '# END RECORD')
        for tport in self.connector.supported_transports:
            mock_iscsiadm.return_value = (sample_output % (tport, tport), '')
            self.assertEqual(tport + '.fake_suffix',
                             self.connector._validate_iface_transport(
                                 tport + '.fake_suffix'))

        mock_iscsiadm.return_value = ("", 'iscsiadm: Could not '
                                      'read iface fake_transport (6)')
        self.assertEqual('default',
                         self.connector._validate_iface_transport(
                             'fake_transport'))

    def test_get_search_path(self):
        search_path = self.connector.get_search_path()
        expected = "/dev/disk/by-path"
        self.assertEqual(expected, search_path)

    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(iscsi.ISCSIConnector, '_get_potential_volume_paths')
    def test_get_volume_paths(self, mock_potential_paths, mock_exists):
        name1 = 'volume-00000001-1'
        vol = {'id': 1, 'name': name1}
        location = '10.0.2.15:3260'
        iqn = 'iqn.2010-10.org.openstack:%s' % name1

        fake_path = ("/dev/disk/by-path/ip-%(ip)s-iscsi-%(iqn)s-lun-%(lun)s" %
                     {'ip': '10.0.2.15', 'iqn': iqn, 'lun': 1})
        fake_devices = [fake_path]
        expected = fake_devices
        mock_potential_paths.return_value = fake_devices

        connection_properties = self.iscsi_connection(vol, [location],
                                                      [iqn])
        volume_paths = self.connector.get_volume_paths(
            connection_properties['data'])
        self.assertEqual(expected, volume_paths)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    def test_discover_mpath_device(self, mock_multipath_device,
                                   mock_multipath_device_path):
        location1 = '10.0.2.15:3260'
        location2 = '[2001:db8::1]:3260'
        name1 = 'volume-00000001-1'
        name2 = 'volume-00000001-2'
        iqn1 = 'iqn.2010-10.org.openstack:%s' % name1
        iqn2 = 'iqn.2010-10.org.openstack:%s' % name2
        fake_multipath_dev = '/dev/mapper/fake-multipath-dev'
        fake_raw_dev = '/dev/disk/by-path/fake-raw-lun'
        vol = {'id': 1, 'name': name1}
        connection_properties = self.iscsi_connection_multipath(
            vol, [location1, location2], [iqn1, iqn2], [1, 2])
        mock_multipath_device_path.return_value = fake_multipath_dev
        mock_multipath_device.return_value = test_connector.FAKE_SCSI_WWN
        (result_path, result_mpath_id) = (
            self.connector_with_multipath._discover_mpath_device(
                test_connector.FAKE_SCSI_WWN,
                connection_properties['data'],
                fake_raw_dev))
        result = {'path': result_path, 'multipath_id': result_mpath_id}
        expected_result = {'path': fake_multipath_dev,
                           'multipath_id': test_connector.FAKE_SCSI_WWN}
        self.assertEqual(expected_result, result)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    @mock.patch.object(os.path, 'realpath')
    def test_discover_mpath_device_by_realpath(self, mock_realpath,
                                               mock_multipath_device,
                                               mock_multipath_device_path):

        FAKE_SCSI_WWN = '1234567890'
        location1 = '10.0.2.15:3260'
        location2 = '[2001:db8::1]:3260'
        name1 = 'volume-00000001-1'
        name2 = 'volume-00000001-2'
        iqn1 = 'iqn.2010-10.org.openstack:%s' % name1
        iqn2 = 'iqn.2010-10.org.openstack:%s' % name2
        fake_multipath_dev = None
        fake_raw_dev = '/dev/disk/by-path/fake-raw-lun'
        vol = {'id': 1, 'name': name1}
        connection_properties = self.iscsi_connection_multipath(
            vol, [location1, location2], [iqn1, iqn2], [1, 2])
        mock_multipath_device_path.return_value = fake_multipath_dev
        mock_multipath_device.return_value = {
            'device': '/dev/mapper/%s' % FAKE_SCSI_WWN}
        mock_realpath.return_value = '/dev/sdvc'
        (result_path, result_mpath_id) = (
            self.connector_with_multipath._discover_mpath_device(
                FAKE_SCSI_WWN,
                connection_properties['data'],
                fake_raw_dev))
        mock_multipath_device.assert_called_with('/dev/sdvc')
        result = {'path': result_path, 'multipath_id': result_mpath_id}
        expected_result = {'path': '/dev/mapper/%s' % FAKE_SCSI_WWN,
                           'multipath_id': FAKE_SCSI_WWN}
        self.assertEqual(expected_result, result)

    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_multipath_volume')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_single_volume')
    def test_connect_volume_mp(self, con_single_mock, con_mp_mock, clean_mock):
        self.connector.use_multipath = True
        res = self.connector.connect_volume(self.CON_PROPS)
        self.assertEqual(con_mp_mock.return_value, res)
        con_single_mock.assert_not_called()
        con_mp_mock.assert_called_once_with(self.CON_PROPS)
        clean_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_multipath_volume')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_single_volume')
    def test_connect_volume_mp_failure(self, con_single_mock, con_mp_mock,
                                       clean_mock):
        self.connector.use_multipath = True
        con_mp_mock.side_effect = exception.BrickException
        self.assertRaises(exception.BrickException,
                          self.connector.connect_volume, self.CON_PROPS)
        con_single_mock.assert_not_called()
        con_mp_mock.assert_called_once_with(self.CON_PROPS)
        clean_mock.assert_called_once_with(self.CON_PROPS, force=True)

    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_multipath_volume')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_single_volume')
    def test_connect_volume_sp(self, con_single_mock, con_mp_mock, clean_mock):
        self.connector.use_multipath = False
        res = self.connector.connect_volume(self.CON_PROPS)
        self.assertEqual(con_single_mock.return_value, res)
        con_mp_mock.assert_not_called()
        con_single_mock.assert_called_once_with(self.CON_PROPS)
        clean_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_multipath_volume')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_single_volume')
    def test_connect_volume_sp_failure(self, con_single_mock, con_mp_mock,
                                       clean_mock):
        self.connector.use_multipath = False
        con_single_mock.side_effect = exception.BrickException
        self.assertRaises(exception.BrickException,
                          self.connector.connect_volume, self.CON_PROPS)
        con_mp_mock.assert_not_called()
        con_single_mock.assert_called_once_with(self.CON_PROPS)
        clean_mock.assert_called_once_with(self.CON_PROPS, force=True)

    def test_discover_iscsi_portals(self):
        location = '10.0.2.15:3260'
        name = 'volume-00000001'
        iqn = 'iqn.2010-10.org.openstack:%s' % name
        vol = {'id': 1, 'name': name}
        auth_method = 'CHAP'
        auth_username = 'fake_chap_username'
        auth_password = 'fake_chap_password'
        discovery_auth_method = 'CHAP'
        discovery_auth_username = 'fake_chap_username'
        discovery_auth_password = 'fake_chap_password'
        connection_properties = self.iscsi_connection_chap(
            vol, location, iqn, auth_method, auth_username, auth_password,
            discovery_auth_method, discovery_auth_username,
            discovery_auth_password)
        self.connector_with_multipath = iscsi.ISCSIConnector(
            None, execute=self.fake_execute, use_multipath=True)

        for transport in ['default', 'iser', 'badTransport']:
            interface = 'iser' if transport == 'iser' else 'default'
            self.mock_object(self.connector_with_multipath, '_get_transport',
                             mock.Mock(return_value=interface))

            self.connector_with_multipath._discover_iscsi_portals(
                connection_properties['data'])

            expected_cmds = [
                'iscsiadm -m discoverydb -t sendtargets -I %(iface)s '
                '-p %(location)s --op update '
                '-n discovery.sendtargets.auth.authmethod -v %(auth_method)s '
                '-n discovery.sendtargets.auth.username -v %(username)s '
                '-n discovery.sendtargets.auth.password -v %(password)s' %
                {'iface': interface, 'location': location,
                 'auth_method': discovery_auth_method,
                 'username': discovery_auth_username,
                 'password': discovery_auth_password},
                'iscsiadm -m node --op show -p %s' % location,
                'iscsiadm -m discoverydb -t sendtargets -I %(iface)s'
                ' -p %(location)s --discover' % {'iface': interface,
                                                 'location': location},
                'iscsiadm -m node --op show -p %s' % location]
            self.assertEqual(expected_cmds, self.cmds)
            # Reset to run with a different transport type
            self.cmds = list()

    @mock.patch.object(iscsi.ISCSIConnector,
                       '_run_iscsiadm_update_discoverydb')
    @mock.patch.object(os.path, 'exists', return_value=True)
    def test_iscsi_portals_with_chap_discovery(
            self, exists, update_discoverydb):
        location = '10.0.2.15:3260'
        name = 'volume-00000001'
        iqn = 'iqn.2010-10.org.openstack:%s' % name
        vol = {'id': 1, 'name': name}
        auth_method = 'CHAP'
        auth_username = 'fake_chap_username'
        auth_password = 'fake_chap_password'
        discovery_auth_method = 'CHAP'
        discovery_auth_username = 'fake_chap_username'
        discovery_auth_password = 'fake_chap_password'
        connection_properties = self.iscsi_connection_chap(
            vol, location, iqn, auth_method, auth_username, auth_password,
            discovery_auth_method, discovery_auth_username,
            discovery_auth_password)
        self.connector_with_multipath = iscsi.ISCSIConnector(
            None, execute=self.fake_execute, use_multipath=True)
        self.cmds = []
        # The first call returns an error code = 6, mocking an empty
        # discovery db. The second one mocks a successful return and the
        # third one a dummy exit code, which will trigger the
        # TargetPortalNotFound exception in connect_volume
        update_discoverydb.side_effect = [
            putils.ProcessExecutionError(None, None, 6),
            ("", ""),
            putils.ProcessExecutionError(None, None, 9)]

        self.connector_with_multipath._discover_iscsi_portals(
            connection_properties['data'])
        update_discoverydb.assert_called_with(connection_properties['data'])

        expected_cmds = [
            'iscsiadm -m discoverydb -t sendtargets -p %s -I default'
            ' --op new' % location,
            'iscsiadm -m node --op show -p %s' % location,
            'iscsiadm -m discoverydb -t sendtargets -I default -p %s'
            ' --discover' % location,
            'iscsiadm -m node --op show -p %s' % location]
        self.assertEqual(expected_cmds, self.cmds)

        self.assertRaises(exception.TargetPortalNotFound,
                          self.connector_with_multipath.connect_volume,
                          connection_properties['data'])

    def test_get_target_portals_from_iscsiadm_output(self):
        connector = self.connector
        test_output = '''10.15.84.19:3260,1 iqn.1992-08.com.netapp:sn.33615311
                         10.15.85.19:3260,2 iqn.1992-08.com.netapp:sn.33615311
                         '''
        res = connector._get_target_portals_from_iscsiadm_output(test_output)
        ips = ['10.15.84.19:3260', '10.15.85.19:3260']
        iqns = ['iqn.1992-08.com.netapp:sn.33615311',
                'iqn.1992-08.com.netapp:sn.33615311']
        expected = (ips, iqns)
        self.assertEqual(expected, res)

    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    def test_disconnect_volume(self, cleanup_mock):
        res = self.connector.disconnect_volume(mock.sentinel.con_props,
                                               mock.sentinel.dev_info,
                                               mock.sentinel.Force,
                                               mock.sentinel.ignore_errors)
        self.assertEqual(cleanup_mock.return_value, res)
        cleanup_mock.assert_called_once_with(
            mock.sentinel.con_props,
            force=mock.sentinel.Force,
            ignore_errors=mock.sentinel.ignore_errors)

    @ddt.data(True, False)
    @mock.patch.object(iscsi.ISCSIConnector, '_get_transport')
    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsiadm_bare')
    def test_get_discoverydb_portals(self, is_iser, iscsiadm_mock,
                                     transport_mock):
        params = {
            'iqn1': self.SINGLE_CON_PROPS['target_iqn'],
            'iqn2': 'iqn.2004-04.com.qnap:ts-831x:iscsi.cinder-2017.9ef',
            'addr': self.SINGLE_CON_PROPS['target_portal'].replace(':', ','),
            'ip1': self.SINGLE_CON_PROPS['target_portal'],
            'ip2': '192.168.1.3:3260',
            'transport': 'iser' if is_iser else 'default',
            'other_transport': 'default' if is_iser else 'iser',
        }

        iscsiadm_mock.return_value = (
            'SENDTARGETS:\n'
            'DiscoveryAddress: 192.168.1.33,3260\n'
            'DiscoveryAddress: %(addr)s\n'
            'Target: %(iqn1)s\n'
            '	Portal: %(ip2)s,1\n'
            '		Iface Name: %(transport)s\n'
            '	Portal: %(ip1)s,1\n'
            '		Iface Name: %(transport)s\n'
            '	Portal: %(ip1)s,1\n'
            '		Iface Name: %(other_transport)s\n'
            'Target: %(iqn2)s\n'
            '	Portal: %(ip2)s,1\n'
            '		Iface Name: %(transport)s\n'
            '	Portal: %(ip1)s,1\n'
            '		Iface Name: %(transport)s\n'
            'DiscoveryAddress: 192.168.1.38,3260\n'
            'iSNS:\n'
            'No targets found.\n'
            'STATIC:\n'
            'No targets found.\n'
            'FIRMWARE:\n'
            'No targets found.\n' % params, None)
        transport_mock.return_value = 'iser' if is_iser else 'non-iser'

        res = self.connector._get_discoverydb_portals(self.SINGLE_CON_PROPS)
        expected = [(params['ip2'], params['iqn1'],
                     self.SINGLE_CON_PROPS['target_lun']),
                    (params['ip1'], params['iqn1'],
                     self.SINGLE_CON_PROPS['target_lun']),
                    (params['ip2'], params['iqn2'],
                     self.SINGLE_CON_PROPS['target_lun']),
                    (params['ip1'], params['iqn2'],
                     self.SINGLE_CON_PROPS['target_lun'])]
        self.assertListEqual(expected, res)
        iscsiadm_mock.assert_called_once_with(
            ['-m', 'discoverydb', '-o', 'show', '-P', 1])
        transport_mock.assert_called_once_with()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_transport', return_value='')
    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsiadm_bare')
    def test_get_discoverydb_portals_error(self, iscsiadm_mock,
                                           transport_mock):
        """DiscoveryAddress is not present."""
        iscsiadm_mock.return_value = (
            'SENDTARGETS:\n'
            'DiscoveryAddress: 192.168.1.33,3260\n'
            'DiscoveryAddress: 192.168.1.38,3260\n'
            'iSNS:\n'
            'No targets found.\n'
            'STATIC:\n'
            'No targets found.\n'
            'FIRMWARE:\n'
            'No targets found.\n', None)

        self.assertRaises(exception.TargetPortalsNotFound,
                          self.connector._get_discoverydb_portals,
                          self.SINGLE_CON_PROPS)
        iscsiadm_mock.assert_called_once_with(
            ['-m', 'discoverydb', '-o', 'show', '-P', 1])
        transport_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_transport', return_value='')
    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsiadm_bare')
    def test_get_discoverydb_portals_error_is_present(self, iscsiadm_mock,
                                                      transport_mock):
        """DiscoveryAddress is present but wrong iterface."""
        params = {
            'iqn': self.SINGLE_CON_PROPS['target_iqn'],
            'addr': self.SINGLE_CON_PROPS['target_portal'].replace(':', ','),
            'ip': self.SINGLE_CON_PROPS['target_portal'],
        }
        iscsiadm_mock.return_value = (
            'SENDTARGETS:\n'
            'DiscoveryAddress: 192.168.1.33,3260\n'
            'DiscoveryAddress: %(addr)s\n'
            'Target: %(iqn)s\n'
            '	Portal: %(ip)s,1\n'
            '		Iface Name: iser\n'
            'DiscoveryAddress: 192.168.1.38,3260\n'
            'iSNS:\n'
            'No targets found.\n'
            'STATIC:\n'
            'No targets found.\n'
            'FIRMWARE:\n'
            'No targets found.\n' % params, None)

        self.assertRaises(exception.TargetPortalsNotFound,
                          self.connector._get_discoverydb_portals,
                          self.SINGLE_CON_PROPS)
        iscsiadm_mock.assert_called_once_with(
            ['-m', 'discoverydb', '-o', 'show', '-P', 1])
        transport_mock.assert_called_once_with()

    @mock.patch.object(iscsi.ISCSIConnector, '_disconnect_connection')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_connection_devices')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'flush_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'remove_connection',
                       return_value=None)
    def test_cleanup_connection(self, remove_mock, flush_mock, con_devs_mock,
                                discon_mock):
        # Return an ordered dicts instead of normal dict for discon_mock.assert
        con_devs_mock.return_value = collections.OrderedDict((
            (('ip1:port1', 'tgt1'), ({'sda'}, set())),
            (('ip2:port2', 'tgt2'), ({'sdb'}, {'sdc'})),
            (('ip3:port3', 'tgt3'), (set(), set()))))

        with mock.patch.object(self.connector,
                               'use_multipath') as use_mp_mock:
            self.connector._cleanup_connection(
                self.CON_PROPS, ips_iqns_luns=mock.sentinel.ips_iqns_luns,
                force=False, ignore_errors=False)

        con_devs_mock.assert_called_once_with(self.CON_PROPS,
                                              mock.sentinel.ips_iqns_luns)
        remove_mock.assert_called_once_with({'sda', 'sdb'}, use_mp_mock,
                                            False, mock.ANY)
        discon_mock.assert_called_once_with(
            self.CON_PROPS,
            [('ip1:port1', 'tgt1'), ('ip3:port3', 'tgt3')],
            False, mock.ANY)
        flush_mock.assert_not_called()

    @mock.patch('os_brick.exception.ExceptionChainer.__nonzero__',
                mock.Mock(return_value=True))
    @mock.patch('os_brick.exception.ExceptionChainer.__bool__',
                mock.Mock(return_value=True))
    @mock.patch.object(iscsi.ISCSIConnector, '_disconnect_connection')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_connection_devices')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'flush_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'remove_connection',
                       return_value=mock.sentinel.mp_name)
    def test_cleanup_connection_force_failure(self, remove_mock, flush_mock,
                                              con_devs_mock, discon_mock):

        # Return an ordered dicts instead of normal dict for discon_mock.assert
        con_devs_mock.return_value = collections.OrderedDict((
            (('ip1:port1', 'tgt1'), ({'sda'}, set())),
            (('ip2:port2', 'tgt2'), ({'sdb'}, {'sdc'})),
            (('ip3:port3', 'tgt3'), (set(), set()))))

        with mock.patch.object(self.connector, 'use_multipath',
                               wraps=True) as use_mp_mock:
            self.assertRaises(exception.ExceptionChainer,
                              self.connector._cleanup_connection,
                              self.CON_PROPS,
                              ips_iqns_luns=mock.sentinel.ips_iqns_luns,
                              force=mock.sentinel.force, ignore_errors=False)

        con_devs_mock.assert_called_once_with(self.CON_PROPS,
                                              mock.sentinel.ips_iqns_luns)
        remove_mock.assert_called_once_with({'sda', 'sdb'}, use_mp_mock,
                                            mock.sentinel.force, mock.ANY)
        discon_mock.assert_called_once_with(
            self.CON_PROPS,
            [('ip1:port1', 'tgt1'), ('ip3:port3', 'tgt3')],
            mock.sentinel.force, mock.ANY)
        flush_mock.assert_called_once_with(mock.sentinel.mp_name)

    def test_cleanup_connection_no_data_discoverydb(self):
        self.connector.use_multipath = True
        with mock.patch.object(self.connector, '_get_discoverydb_portals',
                               side_effect=exception.TargetPortalsNotFound), \
                mock.patch.object(self.connector._linuxscsi,
                                  'remove_connection') as mock_remove:
            # This will not raise and exception
            self.connector._cleanup_connection(self.SINGLE_CON_PROPS)
            mock_remove.assert_not_called()

    @ddt.data({'do_raise': False, 'force': False},
              {'do_raise': True, 'force': True},
              {'do_raise': True, 'force': False})
    @ddt.unpack
    @mock.patch.object(iscsi.ISCSIConnector, '_disconnect_from_iscsi_portal')
    def test_disconnect_connection(self, disconnect_mock, do_raise, force):
        will_raise = do_raise and not force
        actual_call_args = []

        # Since we reuse the copied dictionary on _disconnect_connection
        # changing its values we cannot use mock's assert_has_calls
        def my_disconnect(con_props):
            actual_call_args.append(con_props.copy())
            if do_raise:
                raise exception.ExceptionChainer()

        disconnect_mock.side_effect = my_disconnect

        connections = (('ip1:port1', 'tgt1'), ('ip2:port2', 'tgt2'))
        original_props = self.CON_PROPS.copy()
        exc = exception.ExceptionChainer()
        if will_raise:
            self.assertRaises(exception.ExceptionChainer,
                              self.connector._disconnect_connection,
                              self.CON_PROPS, connections,
                              force=force, exc=exc)
        else:
            self.connector._disconnect_connection(self.CON_PROPS, connections,
                                                  force=force, exc=exc)

        # Passed properties should not be altered by the method call
        self.assertDictEqual(original_props, self.CON_PROPS)
        expected = [original_props.copy(), original_props.copy()]
        for i, (ip, iqn) in enumerate(connections):
            expected[i].update(target_portal=ip, target_iqn=iqn)
        # If we are failing and not forcing we won't make all the alls
        if will_raise:
            expected = expected[:1]
        self.assertListEqual(expected, actual_call_args)
        # No exceptions have been caught by ExceptionChainer context manager
        self.assertEqual(do_raise, bool(exc))

    def test_disconnect_from_iscsi_portal(self):
        self.connector._disconnect_from_iscsi_portal(self.CON_PROPS)
        expected_prefix = ('iscsiadm -m node -T %s -p %s ' %
                           (self.CON_PROPS['target_iqn'],
                            self.CON_PROPS['target_portal']))
        expected = [
            expected_prefix + '--op update -n node.startup -v manual',
            expected_prefix + '--logout',
            expected_prefix + '--op delete',
        ]
        self.assertListEqual(expected, self.cmds)

    def test_iscsiadm_discover_parsing(self):
        # Ensure that parsing iscsiadm discover ignores cruft.

        ips = ["192.168.204.82:3260", "192.168.204.82:3261"]
        iqns = ["iqn.2010-10.org.openstack:volume-"
                "f9b12623-6ce3-4dac-a71f-09ad4249bdd3",
                "iqn.2010-10.org.openstack:volume-"
                "f9b12623-6ce3-4dac-a71f-09ad4249bdd4"]

        # This slight wonkiness brought to you by pep8, as the actual
        # example output runs about 97 chars wide.
        sample_input = """Loading iscsi modules: done
Starting iSCSI initiator service: done
Setting up iSCSI targets: unused
%s %s
%s %s
""" % (ips[0] + ',1', iqns[0], ips[1] + ',1', iqns[1])
        out = self.connector.\
            _get_target_portals_from_iscsiadm_output(sample_input)
        self.assertEqual((ips, iqns), out)

    def test_sanitize_log_run_iscsiadm(self):
        # Tests that the parameters to the _run_iscsiadm function
        # are sanitized for when passwords are logged.
        def fake_debug(*args, **kwargs):
            self.assertIn('node.session.auth.password', args[0])
            self.assertNotIn('scrubme', args[0])

        volume = {'id': 'fake_uuid'}
        connection_info = self.iscsi_connection(volume,
                                                "10.0.2.15:3260",
                                                "fake_iqn")

        iscsi_properties = connection_info['data']
        with mock.patch.object(iscsi.LOG, 'debug',
                               side_effect=fake_debug) as debug_mock:
            self.connector._iscsiadm_update(iscsi_properties,
                                            'node.session.auth.password',
                                            'scrubme')

            # we don't care what the log message is, we just want to make sure
            # our stub method is called which asserts the password is scrubbed
            self.assertTrue(debug_mock.called)

    @mock.patch.object(iscsi.ISCSIConnector, 'get_volume_paths')
    def test_extend_volume_no_path(self, mock_volume_paths):
        mock_volume_paths.return_value = []
        volume = {'id': 'fake_uuid'}
        connection_info = self.iscsi_connection(volume,
                                                "10.0.2.15:3260",
                                                "fake_iqn")

        self.assertRaises(exception.VolumePathsNotFound,
                          self.connector.extend_volume,
                          connection_info['data'])

    @mock.patch.object(linuxscsi.LinuxSCSI, 'extend_volume')
    @mock.patch.object(iscsi.ISCSIConnector, 'get_volume_paths')
    def test_extend_volume(self, mock_volume_paths, mock_scsi_extend):
        fake_new_size = 1024
        mock_volume_paths.return_value = ['/dev/vdx']
        mock_scsi_extend.return_value = fake_new_size
        volume = {'id': 'fake_uuid'}
        connection_info = self.iscsi_connection(volume,
                                                "10.0.2.15:3260",
                                                "fake_iqn")
        new_size = self.connector.extend_volume(connection_info['data'])
        self.assertEqual(fake_new_size, new_size)

    @mock.patch.object(iscsi.LOG, 'info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'extend_volume')
    @mock.patch.object(iscsi.ISCSIConnector, 'get_volume_paths')
    def test_extend_volume_mask_password(self, mock_volume_paths,
                                         mock_scsi_extend,
                                         mock_log_info):
        fake_new_size = 1024
        mock_volume_paths.return_value = ['/dev/vdx']
        mock_scsi_extend.return_value = fake_new_size
        volume = {'id': 'fake_uuid'}
        connection_info = self.iscsi_connection_chap(
            volume, "10.0.2.15:3260", "fake_iqn",
            'CHAP', 'fake_user', 'fake_password',
            'CHAP1', 'fake_user1', 'fake_password1')
        self.connector.extend_volume(connection_info['data'])

        self.assertEqual(2, mock_log_info.call_count)
        self.assertIn("'auth_password': '***'",
                      str(mock_log_info.call_args_list[0]))
        self.assertIn("'discovery_auth_password': '***'",
                      str(mock_log_info.call_args_list[0]))

    @mock.patch.object(iscsi.LOG, 'warning')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'extend_volume')
    @mock.patch.object(iscsi.ISCSIConnector, 'get_volume_paths')
    def test_extend_volume_mask_password_no_paths(self, mock_volume_paths,
                                                  mock_scsi_extend,
                                                  mock_log_warning):
        fake_new_size = 1024
        mock_volume_paths.return_value = []
        mock_scsi_extend.return_value = fake_new_size
        volume = {'id': 'fake_uuid'}
        connection_info = self.iscsi_connection_chap(
            volume, "10.0.2.15:3260", "fake_iqn",
            'CHAP', 'fake_user', 'fake_password',
            'CHAP1', 'fake_user1', 'fake_password1')

        self.assertRaises(exception.VolumePathsNotFound,
                          self.connector.extend_volume,
                          connection_info['data'])

        self.assertEqual(1, mock_log_warning.call_count)
        self.assertIn("'auth_password': '***'",
                      str(mock_log_warning.call_args_list[0]))
        self.assertIn("'discovery_auth_password': '***'",
                      str(mock_log_warning.call_args_list[0]))

    @mock.patch.object(os.path, 'isdir')
    def test_get_all_available_volumes_path_not_dir(self, mock_isdir):
        mock_isdir.return_value = False
        expected = []
        actual = self.connector.get_all_available_volumes()
        self.assertItemsEqual(expected, actual)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_device_path')
    def test_get_potential_paths_mpath(self, get_path_mock):
        self.connector.use_multipath = True
        res = self.connector._get_potential_volume_paths(self.CON_PROPS)
        get_path_mock.assert_called_once_with(self.CON_PROPS)
        self.assertEqual(get_path_mock.return_value, res)
        self.assertEqual([], self.cmds)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions')
    @mock.patch.object(iscsi.ISCSIConnector, '_get_device_path')
    def test_get_potential_paths_single_path(self, get_path_mock,
                                             get_sessions_mock):
        get_path_mock.side_effect = [['path1'], ['path2'], ['path3', 'path4']]
        get_sessions_mock.return_value = [
            'ip1:port1', 'ip2:port2', 'ip3:port3']

        self.connector.use_multipath = False
        res = self.connector._get_potential_volume_paths(self.CON_PROPS)
        self.assertEqual({'path1', 'path2', 'path3', 'path4'}, set(res))
        get_sessions_mock.assert_called_once_with()

    @mock.patch.object(iscsi.ISCSIConnector, '_discover_iscsi_portals')
    def test_get_ips_iqns_luns_with_target_iqns(self, discover_mock):
        res = self.connector._get_ips_iqns_luns(self.CON_PROPS)
        expected = list(self.connector._get_all_targets(self.CON_PROPS))
        self.assertListEqual(expected, res)
        discover_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_discoverydb_portals')
    @mock.patch.object(iscsi.ISCSIConnector, '_discover_iscsi_portals')
    def test_get_ips_iqns_luns_discoverydb(self, discover_mock,
                                           db_portals_mock):
        db_portals_mock.return_value = [('ip1:port1', 'tgt1', '1'),
                                        ('ip2:port2', 'tgt2', '2')]
        res = self.connector._get_ips_iqns_luns(self.SINGLE_CON_PROPS,
                                                discover=False)
        self.assertListEqual(db_portals_mock.return_value, res)
        db_portals_mock.assert_called_once_with(self.SINGLE_CON_PROPS)
        discover_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_discover_iscsi_portals')
    def test_get_ips_iqns_luns_no_target_iqns_share_iqn(self, discover_mock):
        discover_mock.return_value = [('ip1:port1', 'tgt1', '1'),
                                      ('ip1:port1', 'tgt2', '1'),
                                      ('ip2:port2', 'tgt1', '2'),
                                      ('ip2:port2', 'tgt2', '2')]
        res = self.connector._get_ips_iqns_luns(self.SINGLE_CON_PROPS)
        expected = {('ip1:port1', 'tgt1', '1'),
                    ('ip2:port2', 'tgt1', '2')}
        self.assertEqual(expected, set(res))

    @mock.patch.object(iscsi.ISCSIConnector, '_discover_iscsi_portals')
    def test_get_ips_iqns_luns_no_target_iqns_diff_iqn(self, discover_mock):
        discover_mock.return_value = [('ip1:port1', 'tgt1', '1'),
                                      ('ip2:port2', 'tgt2', '2')]
        res = self.connector._get_ips_iqns_luns(self.SINGLE_CON_PROPS)
        self.assertEqual(discover_mock.return_value, res)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full')
    def test_connect_to_iscsi_portal_all_new(self, get_sessions_mock):
        """Connect creating node and session."""
        session = 'session2'
        get_sessions_mock.side_effect = [
            [('tcp:', 'session1', 'ip1:port1', '1', 'tgt')],
            [('tcp:', 'session1', 'ip1:port1', '1', 'tgt'),
             ('tcp:', session, 'ip1:port1', '-1', 'tgt1')]
        ]
        with mock.patch.object(self.connector, '_execute') as exec_mock:
            exec_mock.side_effect = [('', 'error'), ('', None),
                                     ('', None), ('', None),
                                     ('', None)]
            res = self.connector._connect_to_iscsi_portal(self.CON_PROPS)

        # True refers to "manual scans", since the call to update
        # node.session.scan didn't fail they are set to manual
        self.assertEqual((session, True), res)
        prefix = 'iscsiadm -m node -T tgt1 -p ip1:port1'
        expected_cmds = [
            prefix,
            prefix + ' --interface default --op new',
            prefix + ' --op update -n node.session.scan -v manual',
            prefix + ' --login',
            prefix + ' --op update -n node.startup -v automatic'
        ]
        actual_cmds = [' '.join(args[0]) for args in exec_mock.call_args_list]
        self.assertListEqual(expected_cmds, actual_cmds)
        self.assertEqual(2, get_sessions_mock.call_count)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full')
    def test_connect_to_iscsi_portal_all_exists_chap(self, get_sessions_mock):
        """Node and session already exists and we use chap authentication."""
        session = 'session2'
        get_sessions_mock.return_value = [('tcp:', session, 'ip1:port1',
                                           '-1', 'tgt1')]
        con_props = self.CON_PROPS.copy()
        con_props.update(auth_method='CHAP', auth_username='user',
                         auth_password='pwd')
        res = self.connector._connect_to_iscsi_portal(con_props)
        # False refers to "manual scans", so we have manual iscsi scans
        self.assertEqual((session, True), res)
        prefix = 'iscsiadm -m node -T tgt1 -p ip1:port1'
        expected_cmds = [
            prefix,
            prefix + ' --op update -n node.session.scan -v manual',
            prefix + ' --op update -n node.session.auth.authmethod -v CHAP',
            prefix + ' --op update -n node.session.auth.username -v user',
            prefix + ' --op update -n node.session.auth.password -v pwd',
        ]
        self.assertListEqual(expected_cmds, self.cmds)
        get_sessions_mock.assert_called_once_with()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_iscsi_sessions_full')
    def test_connect_to_iscsi_portal_fail_login(self, get_sessions_mock):
        get_sessions_mock.return_value = []
        with mock.patch.object(self.connector, '_execute') as exec_mock:
            exec_mock.side_effect = [('', None), ('', None),
                                     putils.ProcessExecutionError]
            res = self.connector._connect_to_iscsi_portal(self.CON_PROPS)
        self.assertEqual((None, None), res)
        expected_cmds = ['iscsiadm -m node -T tgt1 -p ip1:port1',
                         'iscsiadm -m node -T tgt1 -p ip1:port1 '
                         '--op update -n node.session.scan -v manual',
                         'iscsiadm -m node -T tgt1 -p ip1:port1 --login']
        actual_cmds = [' '.join(args[0]) for args in exec_mock.call_args_list]
        self.assertListEqual(expected_cmds, actual_cmds)
        get_sessions_mock.assert_called_once_with()

    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       side_effect=(None, 'tgt2'))
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    @mock.patch('time.sleep')
    def test_connect_single_volume(self, sleep_mock, cleanup_mock,
                                   connect_mock, get_wwn_mock):
        def my_connect(rescans, props, data):
            if props['target_iqn'] == 'tgt2':
                # Succeed on second call
                data['found_devices'].append('sdz')

        connect_mock.side_effect = my_connect

        res = self.connector._connect_single_volume(self.CON_PROPS)

        expected = {'type': 'block', 'scsi_wwn': 'tgt2', 'path': '/dev/sdz'}
        self.assertEqual(expected, res)
        get_wwn_mock.assert_has_calls([mock.call(['sdz']), mock.call(['sdz'])])
        sleep_mock.assert_called_once_with(1)
        cleanup_mock.assert_called_once_with(
            {'target_lun': 4, 'volume_id': 'vol_id',
             'target_portal': 'ip1:port1', 'target_iqn': 'tgt1'},
            (('ip1:port1', 'tgt1', 4),),
            force=True, ignore_errors=True)

    @staticmethod
    def _get_connect_vol_data():
        return {'stop_connecting': False, 'num_logins': 0, 'failed_logins': 0,
                'stopped_threads': 0, 'found_devices': [],
                'just_added_devices': []}

    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       side_effect=(None, 'tgt2'))
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch.object(iscsi.ISCSIConnector, '_cleanup_connection')
    @mock.patch('time.sleep')
    def test_connect_single_volume_not_found(self, sleep_mock, cleanup_mock,
                                             connect_mock, get_wwn_mock):

        self.assertRaises(exception.VolumeDeviceNotFound,
                          self.connector._connect_single_volume,
                          self.CON_PROPS)

        get_wwn_mock.assert_not_called()

        # Called twice by the retry mechanism
        self.assertEqual(2, sleep_mock.call_count)

        props = list(self.connector._get_all_targets(self.CON_PROPS))
        calls_per_try = [
            mock.call({'target_portal': prop[0], 'target_iqn': prop[1],
                       'target_lun': prop[2], 'volume_id': 'vol_id'},
                      (prop,), force=True, ignore_errors=True)
            for prop in props
        ]
        cleanup_mock.assert_has_calls(calls_per_try * 3)

        data = self._get_connect_vol_data()
        calls_per_try = [mock.call(self.connector.device_scan_attempts,
                                   {'target_portal': prop[0],
                                    'target_iqn': prop[1],
                                    'target_lun': prop[2],
                                    'volume_id': 'vol_id'},
                                   data)
                         for prop in props]
        connect_mock.assert_has_calls(calls_per_try * 3)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
                       side_effect=[None, 'dm-0'])
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       return_value='wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_wwid')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch('time.sleep')
    def test_connect_multipath_volume_all_succeed(self, sleep_mock,
                                                  connect_mock, add_wwid_mock,
                                                  add_path_mock, get_wwn_mock,
                                                  find_dm_mock):
        def my_connect(rescans, props, data):
            devs = {'tgt1': 'sda', 'tgt2': 'sdb', 'tgt3': 'sdc', 'tgt4': 'sdd'}
            data['stopped_threads'] += 1
            data['num_logins'] += 1
            dev = devs[props['target_iqn']]
            data['found_devices'].append(dev)
            data['just_added_devices'].append(dev)

        connect_mock.side_effect = my_connect

        res = self.connector._connect_multipath_volume(self.CON_PROPS)

        expected = {'type': 'block', 'scsi_wwn': 'wwn', 'multipath_id': 'wwn',
                    'path': '/dev/dm-0'}
        self.assertEqual(expected, res)

        self.assertEqual(1, get_wwn_mock.call_count)
        result = list(get_wwn_mock.call_args[0][0])
        result.sort()
        self.assertEqual(['sda', 'sdb', 'sdc', 'sdd'], result)
        add_wwid_mock.assert_called_once_with('wwn')
        self.assertNotEqual(0, add_path_mock.call_count)
        self.assertGreaterEqual(find_dm_mock.call_count, 2)
        self.assertEqual(4, connect_mock.call_count)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
                       side_effect=[None, 'dm-0'])
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       return_value='wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_wwid')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch('time.sleep')
    def test_connect_multipath_volume_all_fail(self, sleep_mock, connect_mock,
                                               add_wwid_mock, add_path_mock,
                                               get_wwn_mock, find_dm_mock):
        def my_connect(rescans, props, data):
            data['stopped_threads'] += 1
            data['failed_logins'] += 1

        connect_mock.side_effect = my_connect

        self.assertRaises(exception.VolumeDeviceNotFound,
                          self.connector._connect_multipath_volume,
                          self.CON_PROPS)

        get_wwn_mock.assert_not_called()
        add_wwid_mock.assert_not_called()
        add_path_mock.assert_not_called()
        find_dm_mock.assert_not_called()
        self.assertEqual(4 * 3, connect_mock.call_count)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
                       side_effect=[None, 'dm-0'])
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       return_value='wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_wwid')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch('time.sleep')
    def test_connect_multipath_volume_some_fail_mp_found(self, sleep_mock,
                                                         connect_mock,
                                                         add_wwid_mock,
                                                         add_path_mock,
                                                         get_wwn_mock,
                                                         find_dm_mock):
        def my_connect(rescans, props, data):
            devs = {'tgt1': '', 'tgt2': 'sdb', 'tgt3': '', 'tgt4': 'sdd'}
            data['stopped_threads'] += 1
            dev = devs[props['target_iqn']]
            if dev:
                data['num_logins'] += 1
                data['found_devices'].append(dev)
                data['just_added_devices'].append(dev)
            else:
                data['failed_logins'] += 1

        connect_mock.side_effect = my_connect

        res = self.connector._connect_multipath_volume(self.CON_PROPS)

        expected = {'type': 'block', 'scsi_wwn': 'wwn', 'multipath_id': 'wwn',
                    'path': '/dev/dm-0'}
        self.assertEqual(expected, res)
        self.assertEqual(1, get_wwn_mock.call_count)
        result = list(get_wwn_mock.call_args[0][0])
        result.sort()
        self.assertEqual(['sdb', 'sdd'], result)
        add_wwid_mock.assert_called_once_with('wwn')
        self.assertNotEqual(0, add_path_mock.call_count)
        self.assertGreaterEqual(find_dm_mock.call_count, 2)
        self.assertEqual(4, connect_mock.call_count)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
                       return_value=None)
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       return_value='wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_wwid')
    @mock.patch.object(iscsi.time, 'time', side_effect=(0, 0, 11, 0))
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch('time.sleep')
    def test_connect_multipath_volume_some_fail_mp_not_found(self, sleep_mock,
                                                             connect_mock,
                                                             time_mock,
                                                             add_wwid_mock,
                                                             add_path_mock,
                                                             get_wwn_mock,
                                                             find_dm_mock):
        def my_connect(rescans, props, data):
            devs = {'tgt1': '', 'tgt2': 'sdb', 'tgt3': '', 'tgt4': 'sdd'}
            data['stopped_threads'] += 1
            dev = devs[props['target_iqn']]
            if dev:
                data['num_logins'] += 1
                data['found_devices'].append(dev)
                data['just_added_devices'].append(dev)
            else:
                data['failed_logins'] += 1

        connect_mock.side_effect = my_connect

        res = self.connector._connect_multipath_volume(self.CON_PROPS)

        expected = [{'type': 'block', 'scsi_wwn': 'wwn', 'path': '/dev/sdb'},
                    {'type': 'block', 'scsi_wwn': 'wwn', 'path': '/dev/sdd'}]
        # It can only be one of the 2
        self.assertIn(res, expected)
        self.assertEqual(1, get_wwn_mock.call_count)
        result = list(get_wwn_mock.call_args[0][0])
        result.sort()
        self.assertEqual(['sdb', 'sdd'], result)
        add_wwid_mock.assert_called_once_with('wwn')
        self.assertNotEqual(0, add_path_mock.call_count)
        self.assertGreaterEqual(find_dm_mock.call_count, 4)
        self.assertEqual(4, connect_mock.call_count)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm',
                       return_value=None)
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_sysfs_wwn',
                       return_value='wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_path')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_add_wwid')
    @mock.patch.object(iscsi.time, 'time', side_effect=(0, 0, 11, 0))
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_vol')
    @mock.patch('time.sleep', mock.Mock())
    def test_connect_multipath_volume_all_loging_not_found(self,
                                                           connect_mock,
                                                           time_mock,
                                                           add_wwid_mock,
                                                           add_path_mock,
                                                           get_wwn_mock,
                                                           find_dm_mock):
        def my_connect(rescans, props, data):
            data['stopped_threads'] += 1
            data['num_logins'] += 1

        connect_mock.side_effect = my_connect

        self.assertRaises(exception.VolumeDeviceNotFound,
                          self.connector._connect_multipath_volume,
                          self.CON_PROPS)

        get_wwn_mock.assert_not_called()
        add_wwid_mock.assert_not_called()
        add_path_mock.assert_not_called()
        find_dm_mock.assert_not_called()
        self.assertEqual(12, connect_mock.call_count)

    @mock.patch('time.sleep', mock.Mock())
    @mock.patch.object(linuxscsi.LinuxSCSI, 'scan_iscsi')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'device_name_by_hctl',
                       return_value='sda')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_to_iscsi_portal')
    def test_connect_vol(self, connect_mock, dev_name_mock, scan_mock):
        lscsi = self.connector._linuxscsi
        data = self._get_connect_vol_data()
        hctl = [mock.sentinel.host, mock.sentinel.channel,
                mock.sentinel.target, mock.sentinel.lun]

        connect_mock.return_value = (mock.sentinel.session, False)

        with mock.patch.object(lscsi, 'get_hctl',
                               side_effect=(None, hctl)) as hctl_mock:
            self.connector._connect_vol(3, self.CON_PROPS, data)

        expected = self._get_connect_vol_data()
        expected.update(num_logins=1, stopped_threads=1,
                        found_devices=['sda'], just_added_devices=['sda'])
        self.assertDictEqual(expected, data)

        connect_mock.assert_called_once_with(self.CON_PROPS)
        hctl_mock.assert_has_calls([mock.call(mock.sentinel.session,
                                              self.CON_PROPS['target_lun']),
                                    mock.call(mock.sentinel.session,
                                              self.CON_PROPS['target_lun'])])

        scan_mock.assert_called_once_with(*hctl)
        dev_name_mock.assert_called_once_with(mock.sentinel.session, hctl)

    @mock.patch.object(iscsi.ISCSIConnector, '_connect_to_iscsi_portal',
                       return_value=(None, False))
    def test_connect_vol_no_session(self, connect_mock):
        data = self._get_connect_vol_data()

        self.connector._connect_vol(3, self.CON_PROPS, data)

        expected = self._get_connect_vol_data()
        expected.update(failed_logins=1, stopped_threads=1)
        self.assertDictEqual(expected, data)

    @mock.patch('time.sleep', mock.Mock())
    @mock.patch.object(linuxscsi.LinuxSCSI, 'scan_iscsi')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'device_name_by_hctl',
                       return_value=None)
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_to_iscsi_portal')
    def test_connect_vol_not_found(self, connect_mock, dev_name_mock,
                                   scan_mock):
        lscsi = self.connector._linuxscsi
        data = self._get_connect_vol_data()
        hctl = [mock.sentinel.host, mock.sentinel.channel,
                mock.sentinel.target, mock.sentinel.lun]

        # True because we are simulating we have manual scans
        connect_mock.return_value = (mock.sentinel.session, True)

        with mock.patch.object(lscsi, 'get_hctl',
                               side_effect=(hctl,)) as hctl_mock:
            self.connector._connect_vol(3, self.CON_PROPS, data)

        expected = self._get_connect_vol_data()
        expected.update(num_logins=1, stopped_threads=1)
        self.assertDictEqual(expected, data)

        hctl_mock.assert_called_once_with(mock.sentinel.session,
                                          self.CON_PROPS['target_lun'])
        # We have 3 scans because on manual mode we also scan on connect
        scan_mock.assert_has_calls([mock.call(*hctl)] * 3)
        dev_name_mock.assert_has_calls(
            [mock.call(mock.sentinel.session, hctl),
             mock.call(mock.sentinel.session, hctl)])

    @mock.patch('time.sleep', mock.Mock())
    @mock.patch.object(linuxscsi.LinuxSCSI, 'scan_iscsi')
    @mock.patch.object(iscsi.ISCSIConnector, '_connect_to_iscsi_portal')
    def test_connect_vol_stop_connecting(self, connect_mock, scan_mock):
        data = self._get_connect_vol_data()

        def device_name_by_hctl(session, hctl):
            data['stop_connecting'] = True
            return None

        lscsi = self.connector._linuxscsi
        hctl = [mock.sentinel.host, mock.sentinel.channel,
                mock.sentinel.target, mock.sentinel.lun]

        connect_mock.return_value = (mock.sentinel.session, False)

        with mock.patch.object(lscsi, 'get_hctl',
                               return_value=hctl) as hctl_mock, \
                mock.patch.object(
                    lscsi, 'device_name_by_hctl',
                    side_effect=device_name_by_hctl) as dev_name_mock:

            self.connector._connect_vol(3, self.CON_PROPS, data)

        expected = self._get_connect_vol_data()
        expected.update(num_logins=1, stopped_threads=1, stop_connecting=True)
        self.assertDictEqual(expected, data)

        hctl_mock.assert_called_once_with(mock.sentinel.session,
                                          self.CON_PROPS['target_lun'])
        scan_mock.assert_not_called()
        dev_name_mock.assert_called_once_with(mock.sentinel.session, hctl)

    @mock.patch.object(iscsi.ISCSIConnector, '_get_device_link')
    def test__get_connect_result(self, get_link_mock):
        props = self.CON_PROPS.copy()
        props['encrypted'] = False
        res = self.connector._get_connect_result(props, 'wwn', ['sda', 'sdb'])
        expected = {'type': 'block', 'scsi_wwn': 'wwn', 'path': '/dev/sda'}
        self.assertDictEqual(expected, res)
        get_link_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_device_link')
    def test__get_connect_result_mpath(self, get_link_mock):
        props = self.CON_PROPS.copy()
        props['encrypted'] = False
        res = self.connector._get_connect_result(props, 'wwn', ['sda', 'sdb'],
                                                 'mpath')
        expected = {'type': 'block', 'scsi_wwn': 'wwn', 'path': '/dev/mpath',
                    'multipath_id': 'wwn'}
        self.assertDictEqual(expected, res)
        get_link_mock.assert_not_called()

    @mock.patch.object(iscsi.ISCSIConnector, '_get_device_link',
                       return_value='/dev/disk/by-id/scsi-wwn')
    def test__get_connect_result_encrypted(self, get_link_mock):
        props = self.CON_PROPS.copy()
        props['encrypted'] = True
        res = self.connector._get_connect_result(props, 'wwn', ['sda', 'sdb'])
        expected = {'type': 'block', 'scsi_wwn': 'wwn',
                    'path': get_link_mock.return_value}
        self.assertDictEqual(expected, res)
        get_link_mock.assert_called_once_with('wwn', '/dev/sda', None)

    @mock.patch('os.path.realpath', return_value='/dev/sda')
    def test__get_device_link(self, realpath_mock):
        symlink = '/dev/disk/by-id/scsi-wwn'
        res = self.connector._get_device_link('wwn', '/dev/sda', None)
        self.assertEqual(symlink, res)
        realpath_mock.assert_called_once_with(symlink)

    @mock.patch('os.path.realpath', return_value='/dev/dm-0')
    def test__get_device_link_multipath(self, realpath_mock):
        symlink = '/dev/disk/by-id/dm-uuid-mpath-wwn'
        res = self.connector._get_device_link('wwn', '/dev/dm-0', 'wwn')
        self.assertEqual(symlink, res)
        realpath_mock.assert_called_once_with(symlink)

    @mock.patch('os.path.realpath', side_effect=('/dev/sdz', '/dev/sdy',
                                                 '/dev/sda', '/dev/sdx'))
    @mock.patch('os.listdir', return_value=['dm-...', 'scsi-wwn', 'scsi-...'])
    def test__get_device_link_check_links(self, listdir_mock, realpath_mock):
        res = self.connector._get_device_link('wwn', '/dev/sda', None)
        self.assertEqual(res, '/dev/disk/by-id/scsi-wwn')
        listdir_mock.assert_called_once_with('/dev/disk/by-id/')
        realpath_mock.assert_has_calls([
            mock.call('/dev/disk/by-id/scsi-wwn'),
            mock.call('/dev/disk/by-id/dm-...'),
            mock.call('/dev/disk/by-id/scsi-wwn')])

    @mock.patch('os.path.realpath', return_value='/dev/sdz')
    @mock.patch('os.listdir', return_value=['dm-...', 'scsi-...'])
    def test__get_device_link_not_found(self, listdir_mock, realpath_mock):
        self.assertRaises(exception.VolumeDeviceNotFound,
                          self.connector._get_device_link,
                          'wwn', '/dev/sda', None)
        listdir_mock.assert_called_once_with('/dev/disk/by-id/')
        realpath_mock.assert_has_calls([
            mock.call('/dev/disk/by-id/scsi-wwn'),
            mock.call('/dev/disk/by-id/dm-...'),
            mock.call('/dev/disk/by-id/scsi-...')])

    @mock.patch.object(iscsi.ISCSIConnector, '_run_iscsiadm_bare')
    def test_get_node_startup_values(self, run_iscsiadm_bare_mock):
        name1 = 'volume-00000001-1'
        name2 = 'volume-00000001-2'
        name3 = 'volume-00000001-3'
        vol = {'id': 1, 'name': name1}
        location = '10.0.2.15:3260'
        iqn1 = 'iqn.2010-10.org.openstack:%s' % name1
        iqn2 = 'iqn.2010-10.org.openstack:%s' % name2
        iqn3 = 'iqn.2010-10.org.openstack:%s' % name3
        connection_properties = self.iscsi_connection(vol, [location], [iqn1])

        node_startup1 = "manual"
        node_startup2 = "automatic"
        node_startup3 = "manual"
        node_values = (
            '# BEGIN RECORD 2.0-873\n'
            'node.name = %s\n'
            'node.tpgt = 1\n'
            'node.startup = %s\n'
            'iface.hwaddress = <empty>\n'
            '# END RECORD\n'
            '# BEGIN RECORD 2.0-873\n'
            'node.name = %s\n'
            'node.tpgt = 1\n'
            'node.startup = %s\n'
            'iface.hwaddress = <empty>\n'
            '# END RECORD\n'
            '# BEGIN RECORD 2.0-873\n'
            'node.name = %s\n'
            'node.tpgt = 1\n'
            'node.startup = %s\n'
            'iface.hwaddress = <empty>\n'
            '# END RECORD\n') % (iqn1, node_startup1, iqn2, node_startup2,
                                 iqn3, node_startup3)
        run_iscsiadm_bare_mock.return_value = (node_values, None)

        node_startups =\
            self.connector._get_node_startup_values(
                connection_properties['data'])
        expected_node_startups = {iqn1: node_startup1, iqn2: node_startup2,
                                  iqn3: node_startup3}
        self.assertEqual(node_startups, expected_node_startups)

    @mock.patch.object(iscsi.ISCSIConnector, '_execute')
    def test_get_node_startup_values_no_nodes(self, exec_mock):
        connection_properties = {'target_portal': 'ip1:port1'}
        no_nodes_output = ''
        no_nodes_err = 'iscsiadm: No records found\n'
        exec_mock.return_value = (no_nodes_output, no_nodes_err)
        res = self.connector._get_node_startup_values(connection_properties)
        self.assertEqual({}, res)
        exec_mock.assert_called_once_with(
            'iscsiadm', '-m', 'node', '--op', 'show', '-p',
            connection_properties['target_portal'],
            root_helper=self.connector._root_helper, run_as_root=True,
            check_exit_code=(0, 21))

    @mock.patch.object(iscsi.ISCSIConnector, '_get_node_startup_values')
    @mock.patch.object(iscsi.ISCSIConnector, '_iscsiadm_update')
    def test_recover_node_startup_values(self, iscsiadm_update_mock,
                                         get_node_startup_values_mock):
        name1 = 'volume-00000001-1'
        name2 = 'volume-00000001-2'
        name3 = 'volume-00000001-3'
        vol = {'id': 1, 'name': name1}
        location = '10.0.2.15:3260'
        iqn1 = 'iqn.2010-10.org.openstack:%s' % name1
        iqn2 = 'iqn.2010-10.org.openstack:%s' % name2
        iqn3 = 'iqn.2010-10.org.openstack:%s' % name3
        connection_properties = self.iscsi_connection(vol, [location], [iqn1])
        recover_connection = self.iscsi_connection(vol, [location], [iqn2])

        node_startup1 = "manual"
        node_startup2 = "automatic"
        node_startup3 = "manual"
        get_node_startup_values_mock.return_value = {iqn1: node_startup1,
                                                     iqn2: node_startup2,
                                                     iqn3: node_startup3}

        old_node_startup_values = {iqn1: node_startup1,
                                   iqn2: "manual",
                                   iqn3: node_startup3}
        self.connector._recover_node_startup_values(
            connection_properties['data'], old_node_startup_values)
        iscsiadm_update_mock.assert_called_once_with(
            recover_connection['data'], "node.startup", "manual")



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from os_win import utilsfactory

from os_brick.initiator.windows import base as win_conn_base
from os_brick.remotefs import windows_remotefs as remotefs
from os_brick import utils


# The Windows SMBFS connector expects to receive VHD/x images stored on SMB
# shares, exposed by the Cinder SMBFS driver.
class WindowsSMBFSConnector(win_conn_base.BaseWindowsConnector):
    def __init__(self, *args, **kwargs):
        super(WindowsSMBFSConnector, self).__init__(*args, **kwargs)
        # If this flag is set, we use the local paths in case of local
        # shares. This is in fact mandatory in some cases, for example
        # for the Hyper-C scenario.
        self._local_path_for_loopback = kwargs.get('local_path_for_loopback',
                                                   True)

        self._expect_raw_disk = kwargs.get('expect_raw_disk', False)
        self._remotefsclient = remotefs.WindowsRemoteFsClient(
            mount_type='smbfs',
            *args, **kwargs)
        self._smbutils = utilsfactory.get_smbutils()
        self._vhdutils = utilsfactory.get_vhdutils()
        self._diskutils = utilsfactory.get_diskutils()

    @staticmethod
    def get_connector_properties(*args, **kwargs):
        # No connector properties updates in this case.
        return {}

    @utils.trace
    def connect_volume(self, connection_properties):
        self.ensure_share_mounted(connection_properties)
        # This will be a virtual disk image path.
        disk_path = self._get_disk_path(connection_properties)

        if self._expect_raw_disk:
            # The caller expects a direct accessible raw disk. We'll
            # mount the image and bring the new disk offline, which will
            # allow direct IO, while ensuring that any partiton residing
            # on it will be unmounted.
            read_only = connection_properties.get('access_mode') == 'ro'
            self._vhdutils.attach_virtual_disk(disk_path, read_only=read_only)
            raw_disk_path = self._vhdutils.get_virtual_disk_physical_path(
                disk_path)
            dev_num = self._diskutils.get_device_number_from_device_name(
                raw_disk_path)
            self._diskutils.set_disk_offline(dev_num)
        else:
            raw_disk_path = None

        device_info = {'type': 'file',
                       'path': raw_disk_path if self._expect_raw_disk
                               else disk_path}
        return device_info

    @utils.trace
    def disconnect_volume(self, connection_properties, device_info=None,
                          force=False, ignore_errors=False):
        export_path = self._get_export_path(connection_properties)

        disk_path = self._get_disk_path(connection_properties)
        # The detach method will silently continue if the disk is
        # not attached.
        self._vhdutils.detach_virtual_disk(disk_path)

        self._remotefsclient.unmount(export_path)

    def _get_export_path(self, connection_properties):
        return connection_properties['export'].replace('/', '\\')

    def _get_disk_path(self, connection_properties):
        # This is expected to be the share address, as an UNC path.
        export_path = self._get_export_path(connection_properties)
        mount_base = self._remotefsclient.get_mount_base()
        use_local_path = (self._local_path_for_loopback and
                          self._smbutils.is_local_share(export_path))

        disk_dir = export_path
        if mount_base:
            # This will be a symlink pointing to either the share
            # path directly or to the local share path, if requested
            # and available.
            disk_dir = self._remotefsclient.get_mount_point(
                export_path)
        elif use_local_path:
            share_name = self._remotefsclient.get_share_name(export_path)
            disk_dir = self._remotefsclient.get_local_share_path(share_name)

        disk_name = connection_properties['name']
        disk_path = os.path.join(disk_dir, disk_name)
        return disk_path

    def get_search_path(self):
        return self._remotefsclient.get_mount_base()

    @utils.trace
    def get_volume_paths(self, connection_properties):
        return [self._get_disk_path(connection_properties)]

    def ensure_share_mounted(self, connection_properties):
        export_path = self._get_export_path(connection_properties)
        mount_options = connection_properties.get('options')
        self._remotefsclient.mount(export_path, mount_options)

    def extend_volume(self, connection_properties):
        raise NotImplementedError



# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Generic linux scsi subsystem and Multipath utilities.

   Note, this is not iSCSI.
"""
import glob
import os
import re
import six

from oslo_concurrency import processutils as putils
from oslo_log import log as logging

from os_brick import exception
from os_brick import executor
from os_brick.privileged import rootwrap as priv_rootwrap
from os_brick import utils

LOG = logging.getLogger(__name__)

MULTIPATH_ERROR_REGEX = re.compile("\w{3} \d+ \d\d:\d\d:\d\d \|.*$")
MULTIPATH_WWID_REGEX = re.compile("\((?P<wwid>.+)\)")
MULTIPATH_DEVICE_ACTIONS = ['unchanged:', 'reject:', 'reload:',
                            'switchpg:', 'rename:', 'create:',
                            'resize:']


class LinuxSCSI(executor.Executor):
    # As found in drivers/scsi/scsi_lib.c
    WWN_TYPES = {'t10.': '1', 'eui.': '2', 'naa.': '3'}

    def echo_scsi_command(self, path, content):
        """Used to echo strings to scsi subsystem."""

        args = ["-a", path]
        kwargs = dict(process_input=content,
                      run_as_root=True,
                      root_helper=self._root_helper)
        self._execute('tee', *args, **kwargs)

    def get_name_from_path(self, path):
        """Translates /dev/disk/by-path/ entry to /dev/sdX."""

        name = os.path.realpath(path)
        if name.startswith("/dev/"):
            return name
        else:
            return None

    def remove_scsi_device(self, device, force=False, exc=None):
        """Removes a scsi device based upon /dev/sdX name."""
        path = "/sys/block/%s/device/delete" % device.replace("/dev/", "")
        if os.path.exists(path):
            exc = exception.ExceptionChainer() if exc is None else exc
            # flush any outstanding IO first
            with exc.context(force, 'Flushing %s failed', device):
                self.flush_device_io(device)

            LOG.debug("Remove SCSI device %(device)s with %(path)s",
                      {'device': device, 'path': path})
            with exc.context(force, 'Removing %s failed', device):
                self.echo_scsi_command(path, "1")

    @utils.retry(exceptions=exception.VolumePathNotRemoved)
    def wait_for_volumes_removal(self, volumes_names):
        """Wait for device paths to be removed from the system."""
        str_names = ', '.join(volumes_names)
        LOG.debug('Checking to see if SCSI volumes %s have been removed.',
                  str_names)
        exist = [volume_name for volume_name in volumes_names
                 if os.path.exists('/dev/' + volume_name)]
        if exist:
            LOG.debug('%s still exist.', ', '.join(exist))
            raise exception.VolumePathNotRemoved(volume_path=exist)
        LOG.debug("SCSI volumes %s have been removed.", str_names)

    def get_device_info(self, device):
        (out, _err) = self._execute('sg_scan', device, run_as_root=True,
                                    root_helper=self._root_helper)
        dev_info = {'device': device, 'host': None,
                    'channel': None, 'id': None, 'lun': None}
        if out:
            line = out.strip()
            line = line.replace(device + ": ", "")
            info = line.split(" ")

            for item in info:
                if '=' in item:
                    pair = item.split('=')
                    dev_info[pair[0]] = pair[1]
                elif 'scsi' in item:
                    dev_info['host'] = item.replace('scsi', '')

        return dev_info

    def get_sysfs_wwn(self, device_names):
        """Return the wwid from sysfs in any of devices in udev format."""
        wwid = self.get_sysfs_wwid(device_names)
        glob_str = '/dev/disk/by-id/scsi-'
        wwn_paths = glob.glob(glob_str + '*')
        # If we don't have multiple designators on page 0x83
        if wwid and glob_str + wwid in wwn_paths:
            return wwid

        # If we have multiple designators follow the symlinks
        for wwn_path in wwn_paths:
            try:
                if os.path.islink(wwn_path) and os.stat(wwn_path):
                    path = os.path.realpath(wwn_path)
                    if path.startswith('/dev/') and path[5:] in device_names:
                        return wwn_path[len(glob_str):]
            except OSError:
                continue
        return ''

    def get_sysfs_wwid(self, device_names):
        """Return the wwid from sysfs in any of devices in udev format."""
        for device_name in device_names:
            try:
                with open('/sys/block/%s/device/wwid' % device_name) as f:
                    wwid = f.read().strip()
            except IOError:
                continue
            # The sysfs wwid has the wwn type in string format as a prefix,
            # but udev uses its numerical representation as returned by
            # scsi_id's page 0x83, so we need to map it
            udev_wwid = self.WWN_TYPES.get(wwid[:4], '8') + wwid[4:]
            return udev_wwid
        return ''

    def get_scsi_wwn(self, path):
        """Read the WWN from page 0x83 value for a SCSI device."""

        (out, _err) = self._execute('/lib/udev/scsi_id', '--page', '0x83',
                                    '--whitelisted', path,
                                    run_as_root=True,
                                    root_helper=self._root_helper)
        return out.strip()

    @staticmethod
    def is_multipath_running(enforce_multipath, root_helper, execute=None):
        try:
            if execute is None:
                execute = priv_rootwrap.execute
            execute('multipathd', 'show', 'status',
                    run_as_root=True, root_helper=root_helper)
        except putils.ProcessExecutionError as err:
            LOG.error('multipathd is not running: exit code %(err)s',
                      {'err': err.exit_code})
            if enforce_multipath:
                raise
            return False

        return True

    def get_dm_name(self, dm):
        """Get the Device map name given the device name of the dm on sysfs.

        :param dm: Device map name as seen in sysfs. ie: 'dm-0'
        :returns: String with the name, or empty string if not available.
                  ie: '36e843b658476b7ed5bc1d4d10d9b1fde'
        """
        try:
            with open('/sys/block/' + dm + '/dm/name') as f:
                return f.read().strip()
        except IOError:
            return ''

    def find_sysfs_multipath_dm(self, device_names):
        """Find the dm device name given a list of device names

        :param device_names: Iterable with device names, not paths. ie: ['sda']
        :returns: String with the dm name or None if not found. ie: 'dm-0'
        """
        glob_str = '/sys/block/%s/holders/dm-*'
        for dev_name in device_names:
            dms = glob.glob(glob_str % dev_name)
            if dms:
                __, device_name, __, dm = dms[0].rsplit('/', 3)
                return dm
        return None

    def remove_connection(self, devices_names, is_multipath, force=False,
                          exc=None):
        """Remove LUNs and multipath associated with devices names.

        :param devices_names: Iterable with real device names ('sda', 'sdb')
        :param is_multipath: Whether this is a multipath connection or not
        :param force: Whether to forcefully disconnect even if flush fails.
        :param exc: ExceptionChainer where to add exceptions if forcing
        :returns: Multipath device map name if found and not flushed
        """
        if not devices_names:
            return
        multipath_name = None
        exc = exception.ExceptionChainer() if exc is None else exc
        LOG.debug('Removing %(type)s devices %(devices)s',
                  {'type': 'multipathed' if is_multipath else 'single pathed',
                   'devices': ', '.join(devices_names)})

        if is_multipath:
            multipath_dm = self.find_sysfs_multipath_dm(devices_names)
            multipath_name = multipath_dm and self.get_dm_name(multipath_dm)
            if multipath_name:
                with exc.context(force, 'Flushing %s failed', multipath_name):
                    self.flush_multipath_device(multipath_name)
                    multipath_name = None

        for device_name in devices_names:
            self.remove_scsi_device('/dev/' + device_name, force, exc)

        # Wait until the symlinks are removed
        with exc.context(force, 'Some devices remain from %s', devices_names):
            try:
                self.wait_for_volumes_removal(devices_names)
            finally:
                # Since we use /dev/disk/by-id/scsi- links to get the wwn we
                # must ensure they are always removed.
                self._remove_scsi_symlinks(devices_names)
        return multipath_name

    def _remove_scsi_symlinks(self, devices_names):
        devices = ['/dev/' + dev for dev in devices_names]
        links = glob.glob('/dev/disk/by-id/scsi-*')
        unlink = []
        for link in links:
            try:
                if os.path.realpath(link) in devices:
                    unlink.append(link)
            except OSError:
                # A race condition in Python's posixpath:realpath just occurred
                # so we can ignore it because the file was just removed between
                # a check if file exists and a call to os.readlink
                continue

        if unlink:
            priv_rootwrap.unlink_root(no_errors=True, *unlink)

    def flush_device_io(self, device):
        """This is used to flush any remaining IO in the buffers."""
        if os.path.exists(device):
            try:
                # NOTE(geguileo): With 30% connection error rates flush can get
                # stuck, set timeout to prevent it from hanging here forever.
                # Retry twice after 20 and 40 seconds.
                LOG.debug("Flushing IO for device %s", device)
                self._execute('blockdev', '--flushbufs', device,
                              run_as_root=True, attempts=3, timeout=300,
                              interval=10, root_helper=self._root_helper)
            except putils.ProcessExecutionError as exc:
                LOG.warning("Failed to flush IO buffers prior to removing "
                            "device: %(code)s", {'code': exc.exit_code})
                raise

    def flush_multipath_device(self, device_map_name):
        LOG.debug("Flush multipath device %s", device_map_name)
        # NOTE(geguileo): With 30% connection error rates flush can get stuck,
        # set timeout to prevent it from hanging here forever.  Retry twice
        # after 20 and 40 seconds.
        self._execute('multipath', '-f', device_map_name, run_as_root=True,
                      attempts=3, timeout=300, interval=10,
                      root_helper=self._root_helper)

    @utils.retry(exceptions=exception.VolumeDeviceNotFound)
    def wait_for_path(self, volume_path):
        """Wait for a path to show up."""
        LOG.debug("Checking to see if %s exists yet.",
                  volume_path)
        if not os.path.exists(volume_path):
            LOG.debug("%(path)s doesn't exists yet.", {'path': volume_path})
            raise exception.VolumeDeviceNotFound(
                device=volume_path)
        else:
            LOG.debug("%s has shown up.", volume_path)

    @utils.retry(exceptions=exception.BlockDeviceReadOnly, retries=5)
    def wait_for_rw(self, wwn, device_path):
        """Wait for block device to be Read-Write."""
        LOG.debug("Checking to see if %s is read-only.",
                  device_path)
        out, info = self._execute('lsblk', '-o', 'NAME,RO', '-l', '-n')
        LOG.debug("lsblk output: %s", out)
        blkdevs = out.splitlines()
        for blkdev in blkdevs:
            # Entries might look like:
            #
            #   "3624a93709a738ed78583fd120013902b (dm-1)  1"
            #
            # or
            #
            #   "sdd                                       0"
            #
            # We are looking for the first and last part of them. For FC
            # multipath devices the name is in the format of '<WWN> (dm-<ID>)'
            blkdev_parts = blkdev.split(' ')
            ro = blkdev_parts[-1]
            name = blkdev_parts[0]

            # We must validate that all pieces of the dm-# device are rw,
            # if some are still ro it can cause problems.
            if wwn in name and int(ro) == 1:
                LOG.debug("Block device %s is read-only", device_path)
                self._execute('multipath', '-r', check_exit_code=[0, 1, 21],
                              run_as_root=True, root_helper=self._root_helper)
                raise exception.BlockDeviceReadOnly(
                    device=device_path)
        else:
            LOG.debug("Block device %s is not read-only.", device_path)

    def find_multipath_device_path(self, wwn):
        """Look for the multipath device file for a volume WWN.

        Multipath devices can show up in several places on
        a linux system.

        1) When multipath friendly names are ON:
            a device file will show up in
            /dev/disk/by-id/dm-uuid-mpath-<WWN>
            /dev/disk/by-id/dm-name-mpath<N>
            /dev/disk/by-id/scsi-mpath<N>
            /dev/mapper/mpath<N>

        2) When multipath friendly names are OFF:
            /dev/disk/by-id/dm-uuid-mpath-<WWN>
            /dev/disk/by-id/scsi-<WWN>
            /dev/mapper/<WWN>

        """
        LOG.info("Find Multipath device file for volume WWN %(wwn)s",
                 {'wwn': wwn})
        # First look for the common path
        wwn_dict = {'wwn': wwn}
        path = "/dev/disk/by-id/dm-uuid-mpath-%(wwn)s" % wwn_dict
        try:
            self.wait_for_path(path)
            return path
        except exception.VolumeDeviceNotFound:
            pass

        # for some reason the common path wasn't found
        # lets try the dev mapper path
        path = "/dev/mapper/%(wwn)s" % wwn_dict
        try:
            self.wait_for_path(path)
            return path
        except exception.VolumeDeviceNotFound:
            pass

        # couldn't find a path
        LOG.warning("couldn't find a valid multipath device path for "
                    "%(wwn)s", wwn_dict)
        return None

    def find_multipath_device(self, device):
        """Discover multipath devices for a mpath device.

           This uses the slow multipath -l command to find a
           multipath device description, then screen scrapes
           the output to discover the multipath device name
           and it's devices.

        """

        mdev = None
        devices = []
        out = None
        try:
            (out, _err) = self._execute('multipath', '-l', device,
                                        run_as_root=True,
                                        root_helper=self._root_helper)
        except putils.ProcessExecutionError as exc:
            LOG.warning("multipath call failed exit %(code)s",
                        {'code': exc.exit_code})
            raise exception.CommandExecutionFailed(
                cmd='multipath -l %s' % device)

        if out:
            lines = out.strip()
            lines = lines.split("\n")
            lines = [line for line in lines
                     if not re.match(MULTIPATH_ERROR_REGEX, line)
                     and len(line)]
            if lines:

                mdev_name = lines[0].split(" ")[0]

                if mdev_name in MULTIPATH_DEVICE_ACTIONS:
                    mdev_name = lines[0].split(" ")[1]

                mdev = '/dev/mapper/%s' % mdev_name

                # Confirm that the device is present.
                try:
                    os.stat(mdev)
                except OSError:
                    LOG.warning("Couldn't find multipath device %s",
                                mdev)
                    return None

                wwid_search = MULTIPATH_WWID_REGEX.search(lines[0])
                if wwid_search is not None:
                    mdev_id = wwid_search.group('wwid')
                else:
                    mdev_id = mdev_name

                LOG.debug("Found multipath device = %(mdev)s",
                          {'mdev': mdev})
                device_lines = lines[3:]
                for dev_line in device_lines:
                    if dev_line.find("policy") != -1:
                        continue

                    dev_line = dev_line.lstrip(' |-`')
                    dev_info = dev_line.split()
                    address = dev_info[0].split(":")

                    dev = {'device': '/dev/%s' % dev_info[1],
                           'host': address[0], 'channel': address[1],
                           'id': address[2], 'lun': address[3]
                           }

                    devices.append(dev)

        if mdev is not None:
            info = {"device": mdev,
                    "id": mdev_id,
                    "name": mdev_name,
                    "devices": devices}
            return info
        return None

    def get_device_size(self, device):
        """Get the size in bytes of a volume."""
        (out, _err) = self._execute('blockdev', '--getsize64',
                                    device, run_as_root=True,
                                    root_helper=self._root_helper)
        var = six.text_type(out.strip())
        if var.isnumeric():
            return int(var)
        else:
            return None

    def multipath_reconfigure(self):
        """Issue a multipathd reconfigure.

        When attachments come and go, the multipathd seems
        to get lost and not see the maps.  This causes
        resize map to fail 100%.  To overcome this we have
        to issue a reconfigure prior to resize map.
        """
        (out, _err) = self._execute('multipathd', 'reconfigure',
                                    run_as_root=True,
                                    root_helper=self._root_helper)
        return out

    def multipath_resize_map(self, mpath_id):
        """Issue a multipath resize map on device.

        This forces the multipath daemon to update it's
        size information a particular multipath device.
        """
        (out, _err) = self._execute('multipathd', 'resize', 'map', mpath_id,
                                    run_as_root=True,
                                    root_helper=self._root_helper)
        return out

    def extend_volume(self, volume_paths):
        """Signal the SCSI subsystem to test for volume resize.

        This function tries to signal the local system's kernel
        that an already attached volume might have been resized.
        """
        LOG.debug("extend volume %s", volume_paths)

        for volume_path in volume_paths:
            device = self.get_device_info(volume_path)
            LOG.debug("Volume device info = %s", device)
            device_id = ("%(host)s:%(channel)s:%(id)s:%(lun)s" %
                         {'host': device['host'],
                          'channel': device['channel'],
                          'id': device['id'],
                          'lun': device['lun']})

            scsi_path = ("/sys/bus/scsi/drivers/sd/%(device_id)s" %
                         {'device_id': device_id})

            size = self.get_device_size(volume_path)
            LOG.debug("Starting size: %s", size)

            # now issue the device rescan
            rescan_path = "%(scsi_path)s/rescan" % {'scsi_path': scsi_path}
            self.echo_scsi_command(rescan_path, "1")
            new_size = self.get_device_size(volume_path)
            LOG.debug("volume size after scsi device rescan %s", new_size)

        scsi_wwn = self.get_scsi_wwn(volume_paths[0])
        mpath_device = self.find_multipath_device_path(scsi_wwn)
        if mpath_device:
            # Force a reconfigure so that resize works
            self.multipath_reconfigure()

            size = self.get_device_size(mpath_device)
            LOG.info("mpath(%(device)s) current size %(size)s",
                     {'device': mpath_device, 'size': size})
            result = self.multipath_resize_map(scsi_wwn)
            if 'fail' in result:
                LOG.error("Multipathd failed to update the size mapping of "
                          "multipath device %(scsi_wwn)s volume %(volume)s",
                          {'scsi_wwn': scsi_wwn, 'volume': volume_paths})
                return None

            new_size = self.get_device_size(mpath_device)
            LOG.info("mpath(%(device)s) new size %(size)s",
                     {'device': mpath_device, 'size': new_size})

        return new_size

    def process_lun_id(self, lun_ids):
        if isinstance(lun_ids, list):
            processed = []
            for x in lun_ids:
                x = self._format_lun_id(x)
                processed.append(x)
        else:
            processed = self._format_lun_id(lun_ids)
        return processed

    def _format_lun_id(self, lun_id):
        # make sure lun_id is an int
        lun_id = int(lun_id)
        if lun_id < 256:
            return lun_id
        else:
            return ("0x%04x%04x00000000" %
                    (lun_id & 0xffff, lun_id >> 16 & 0xffff))

    def get_hctl(self, session, lun):
        """Given an iSCSI session return the host, channel, target, and lun."""
        glob_str = '/sys/class/iscsi_host/host*/device/session' + session
        paths = glob.glob(glob_str + '/target*')
        if paths:
            __, channel, target = os.path.split(paths[0])[1].split(':')
        # Check if we can get the host
        else:
            target = channel = '-'
            paths = glob.glob(glob_str)

        if not paths:
            LOG.debug('No hctl found on session %s with lun %s', session, lun)
            return None

        # Extract the host number from the path
        host = paths[0][26:paths[0].index('/', 26)]
        res = (host, channel, target, lun)
        LOG.debug('HCTL %s found on session %s with lun %s', res, session, lun)
        return res

    def device_name_by_hctl(self, session, hctl):
        """Find the device name given a session and the hctl.

        :param session: A string with the session number
        "param hctl: An iterable with the host, channel, target, and lun as
                     passed to scan.  ie: ('5', '-', '-', '0')
        """
        if '-' in hctl:
            hctl = ['*' if x == '-' else x for x in hctl]
        path = ('/sys/class/scsi_host/host%(h)s/device/session%(s)s/target'
                '%(h)s:%(c)s:%(t)s/%(h)s:%(c)s:%(t)s:%(l)s/block/*' %
                {'h': hctl[0], 'c': hctl[1], 't': hctl[2], 'l': hctl[3],
                 's': session})
        # Sort devices and return the first so we don't return a partition
        devices = sorted(glob.glob(path))
        device = os.path.split(devices[0])[1] if devices else None
        LOG.debug('Searching for a device in session %s and hctl %s yield: %s',
                  session, hctl, device)
        return device

    def scan_iscsi(self, host, channel='-', target='-', lun='-'):
        """Send an iSCSI scan request given the host and optionally the ctl."""
        LOG.debug('Scanning host %(host)s c: %(channel)s, '
                  't: %(target)s, l: %(lun)s)',
                  {'host': host, 'channel': channel,
                   'target': target, 'lun': lun})
        self.echo_scsi_command('/sys/class/scsi_host/host%s/scan' % host,
                               '%(c)s %(t)s %(l)s' % {'c': channel,
                                                      't': target,
                                                      'l': lun})

    def multipath_add_wwid(self, wwid):
        """Add a wwid to the list of know multipath wwids.

        This has the effect of multipathd being willing to create a dm for a
        multipath even when there's only 1 device.
        """
        out, err = self._execute('multipath', '-a', wwid,
                                 run_as_root=True,
                                 check_exit_code=False,
                                 root_helper=self._root_helper)
        return out.strip() == "wwid '" + wwid + "' added"

    def multipath_add_path(self, realpath):
        """Add a path to multipathd for monitoring.

        This has the effect of multipathd checking an already checked device
        for multipath.

        Together with `multipath_add_wwid` we can create a multipath when
        there's only 1 path.
        """
        stdout, stderr = self._execute('multipathd', 'add', 'path', realpath,
                                       run_as_root=True, timeout=5,
                                       check_exit_code=False,
                                       root_helper=self._root_helper)
        return stdout.strip() == 'ok'



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import ddt
import mock

from os_brick.initiator.windows import smbfs
from os_brick.remotefs import windows_remotefs
from os_brick.tests.windows import test_base


@ddt.ddt
class WindowsSMBFSConnectorTestCase(test_base.WindowsConnectorTestBase):
    def setUp(self):
        super(WindowsSMBFSConnectorTestCase, self).setUp()

        self._load_connector()

    @mock.patch.object(windows_remotefs, 'WindowsRemoteFsClient')
    def _load_connector(self, mock_remotefs_cls, *args, **kwargs):
        self._connector = smbfs.WindowsSMBFSConnector(*args, **kwargs)
        self._remotefs = mock_remotefs_cls.return_value

        self._vhdutils = self._connector._vhdutils
        self._diskutils = self._connector._diskutils

    @mock.patch.object(smbfs.WindowsSMBFSConnector, '_get_disk_path')
    @mock.patch.object(smbfs.WindowsSMBFSConnector, 'ensure_share_mounted')
    def test_connect_volume(self, mock_ensure_mounted,
                            mock_get_disk_path):
        device_info = self._connector.connect_volume(mock.sentinel.conn_props)
        expected_info = dict(type='file',
                             path=mock_get_disk_path.return_value)

        self.assertEqual(expected_info, device_info)
        mock_ensure_mounted.assert_called_once_with(mock.sentinel.conn_props)
        mock_get_disk_path.assert_called_once_with(mock.sentinel.conn_props)

    @ddt.data(True, False)
    @mock.patch.object(smbfs.WindowsSMBFSConnector, '_get_disk_path')
    @mock.patch.object(smbfs.WindowsSMBFSConnector, 'ensure_share_mounted')
    def test_connect_and_mount_volume(self, read_only,
                                      mock_ensure_mounted,
                                      mock_get_disk_path):
        self._load_connector(expect_raw_disk=True)

        fake_conn_props = dict(access_mode='ro' if read_only else 'rw')
        self._vhdutils.get_virtual_disk_physical_path.return_value = (
            mock.sentinel.raw_disk_path)
        mock_get_disk_path.return_value = mock.sentinel.image_path

        device_info = self._connector.connect_volume(fake_conn_props)

        expected_info = dict(type='file',
                             path=mock.sentinel.raw_disk_path)

        self.assertEqual(expected_info, device_info)
        self._vhdutils.attach_virtual_disk.assert_called_once_with(
            mock.sentinel.image_path,
            read_only=read_only)
        self._vhdutils.get_virtual_disk_physical_path.assert_called_once_with(
            mock.sentinel.image_path)
        get_dev_num = self._diskutils.get_device_number_from_device_name
        get_dev_num.assert_called_once_with(mock.sentinel.raw_disk_path)
        self._diskutils.set_disk_offline.assert_called_once_with(
            get_dev_num.return_value)

    @mock.patch.object(smbfs.WindowsSMBFSConnector, '_get_disk_path')
    @mock.patch.object(smbfs.WindowsSMBFSConnector, '_get_export_path')
    def test_disconnect_volume(self, mock_get_export_path,
                               mock_get_disk_path):
        self._connector.disconnect_volume(mock.sentinel.conn_props,
                                          mock.sentinel.dev_info)

        mock_get_disk_path.assert_called_once_with(
            mock.sentinel.conn_props)
        self._vhdutils.detach_virtual_disk.assert_called_once_with(
            mock_get_disk_path.return_value)
        self._remotefs.unmount.assert_called_once_with(
            mock_get_export_path.return_value)
        mock_get_export_path.assert_called_once_with(mock.sentinel.conn_props)

    def test_get_export_path(self):
        fake_export = '//ip/share'
        fake_conn_props = dict(export=fake_export)

        expected_export = fake_export.replace('/', '\\')
        export_path = self._connector._get_export_path(fake_conn_props)
        self.assertEqual(expected_export, export_path)

    @ddt.data({},
              {'mount_base': mock.sentinel.mount_base},
              {'is_local_share': True},
              {'is_local_share': True,
               'local_path_for_loopbk': True})
    @ddt.unpack
    def test_get_disk_path(self, mount_base=None,
                           local_path_for_loopbk=False,
                           is_local_share=False):
        fake_mount_point = r'C:\\fake_mount_point'
        fake_share_name = 'fake_share'
        fake_local_share_path = 'C:\\%s' % fake_share_name
        fake_export_path = '\\\\host\\%s' % fake_share_name
        fake_disk_name = 'fake_disk.vhdx'
        fake_conn_props = dict(name=fake_disk_name,
                               export=fake_export_path)

        self._remotefs.get_mount_base.return_value = mount_base
        self._remotefs.get_mount_point.return_value = fake_mount_point
        self._remotefs.get_local_share_path.return_value = (
            fake_local_share_path)
        self._remotefs.get_share_name.return_value = fake_share_name
        self._connector._local_path_for_loopback = local_path_for_loopbk
        self._connector._smbutils.is_local_share.return_value = is_local_share

        expecting_local = local_path_for_loopbk and is_local_share

        if mount_base:
            expected_export_path = fake_mount_point
        elif expecting_local:
            # In this case, we expect the local share export path to be
            # used directly.
            expected_export_path = fake_local_share_path
        else:
            expected_export_path = fake_export_path
        expected_disk_path = os.path.join(expected_export_path,
                                          fake_disk_name)

        disk_path = self._connector._get_disk_path(fake_conn_props)
        self.assertEqual(expected_disk_path, disk_path)

        if mount_base:
            self._remotefs.get_mount_point.assert_called_once_with(
                fake_export_path)
        elif expecting_local:
            self._connector._smbutils.is_local_share.assert_called_once_with(
                fake_export_path)
            self._remotefs.get_share_name.assert_called_once_with(
                fake_export_path)
            self._remotefs.get_local_share_path.assert_called_once_with(
                fake_share_name)

    def test_get_search_path(self):
        search_path = self._connector.get_search_path()
        self.assertEqual(search_path,
                         self._remotefs.get_mount_base.return_value)

    @mock.patch.object(smbfs.WindowsSMBFSConnector, '_get_disk_path')
    def test_volume_paths(self, mock_get_disk_path):
        expected_paths = [mock_get_disk_path.return_value]
        volume_paths = self._connector.get_volume_paths(
            mock.sentinel.conn_props)

        self.assertEqual(expected_paths, volume_paths)
        mock_get_disk_path.assert_called_once_with(
            mock.sentinel.conn_props)

    @mock.patch.object(smbfs.WindowsSMBFSConnector, '_get_export_path')
    def test_ensure_share_mounted(self, mock_get_export_path):
        fake_conn_props = dict(options=mock.sentinel.mount_opts)

        self._connector.ensure_share_mounted(fake_conn_props)
        self._remotefs.mount.assert_called_once_with(
            mock_get_export_path.return_value,
            mock.sentinel.mount_opts)



# Copyright 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from os_brick.initiator.connectors import base
from os_brick.initiator.connectors import base_iscsi


class FakeConnector(base.BaseLinuxConnector):

    fake_path = '/dev/vdFAKE'

    def connect_volume(self, connection_properties):
        fake_device_info = {'type': 'fake',
                            'path': self.fake_path}
        return fake_device_info

    def disconnect_volume(self, connection_properties, device_info,
                          force=False, ignore_errors=False):
        pass

    def get_volume_paths(self, connection_properties):
        return [self.fake_path]

    def get_search_path(self):
        return '/dev/disk/by-path'

    def extend_volume(self, connection_properties):
        return None

    def get_all_available_volumes(self, connection_properties=None):
        return ['/dev/disk/by-path/fake-volume-1',
                '/dev/disk/by-path/fake-volume-X']


class FakeBaseISCSIConnector(FakeConnector, base_iscsi.BaseISCSIConnector):
    pass



# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
import os
import six

from os_brick import exception
from os_brick.initiator.connectors import base
from os_brick.initiator.connectors import fibre_channel
from os_brick.initiator import linuxfc
from os_brick.initiator import linuxscsi
from os_brick.tests.initiator import test_connector


@ddt.ddt
class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):

    def setUp(self):
        super(FibreChannelConnectorTestCase, self).setUp()
        self.connector = fibre_channel.FibreChannelConnector(
            None, execute=self.fake_execute, use_multipath=False)
        self.assertIsNotNone(self.connector)
        self.assertIsNotNone(self.connector._linuxfc)
        self.assertIsNotNone(self.connector._linuxscsi)

    def fake_get_fc_hbas(self):
        return [{'ClassDevice': 'host1',
                 'ClassDevicePath': '/sys/devices/pci0000:00/0000:00:03.0'
                                    '/0000:05:00.2/host1/fc_host/host1',
                 'dev_loss_tmo': '30',
                 'fabric_name': '0x1000000533f55566',
                 'issue_lip': '<store method only>',
                 'max_npiv_vports': '255',
                 'maxframe_size': '2048 bytes',
                 'node_name': '0x200010604b019419',
                 'npiv_vports_inuse': '0',
                 'port_id': '0x680409',
                 'port_name': '0x100010604b019419',
                 'port_state': 'Online',
                 'port_type': 'NPort (fabric via point-to-point)',
                 'speed': '10 Gbit',
                 'supported_classes': 'Class 3',
                 'supported_speeds': '10 Gbit',
                 'symbolic_name': 'Emulex 554M FV4.0.493.0 DV8.3.27',
                 'tgtid_bind_type': 'wwpn (World Wide Port Name)',
                 'uevent': None,
                 'vport_create': '<store method only>',
                 'vport_delete': '<store method only>'}]

    def fake_get_fc_hbas_info(self):
        hbas = self.fake_get_fc_hbas()
        info = [{'port_name': hbas[0]['port_name'].replace('0x', ''),
                 'node_name': hbas[0]['node_name'].replace('0x', ''),
                 'host_device': hbas[0]['ClassDevice'],
                 'device_path': hbas[0]['ClassDevicePath']}]
        return info

    def fibrechan_connection(self, volume, location, wwn, lun=1):
        return {'driver_volume_type': 'fibrechan',
                'data': {
                    'volume_id': volume['id'],
                    'target_portal': location,
                    'target_wwn': wwn,
                    'target_lun': lun,
                }}

    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    def test_get_connector_properties(self, mock_hbas):
        mock_hbas.return_value = self.fake_get_fc_hbas()
        multipath = True
        enforce_multipath = True
        props = fibre_channel.FibreChannelConnector.get_connector_properties(
            'sudo', multipath=multipath,
            enforce_multipath=enforce_multipath)

        hbas = self.fake_get_fc_hbas()
        expected_props = {'wwpns': [hbas[0]['port_name'].replace('0x', '')],
                          'wwnns': [hbas[0]['node_name'].replace('0x', '')]}
        self.assertEqual(expected_props, props)

    def test_get_search_path(self):
        search_path = self.connector.get_search_path()
        expected = "/dev/disk/by-path"
        self.assertEqual(expected, search_path)

    def test_get_pci_num(self):
        hba = {'device_path': "/sys/devices/pci0000:00/0000:00:03.0"
                              "/0000:05:00.3/host2/fc_host/host2"}
        pci_num = self.connector._get_pci_num(hba)
        self.assertEqual("0000:05:00.3", pci_num)

        hba = {'device_path': "/sys/devices/pci0000:00/0000:00:03.0"
                              "/0000:05:00.3/0000:06:00.6/host2/fc_host/host2"}
        pci_num = self.connector._get_pci_num(hba)
        self.assertEqual("0000:06:00.6", pci_num)

        hba = {'device_path': "/sys/devices/pci0000:20/0000:20:03.0"
                              "/0000:21:00.2/net/ens2f2/ctlr_2/host3"
                              "/fc_host/host3"}
        pci_num = self.connector._get_pci_num(hba)
        self.assertEqual("0000:21:00.2", pci_num)

    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    def test_get_volume_paths(self, fake_fc_hbas_info,
                              fake_fc_hbas, fake_exists):
        fake_fc_hbas.side_effect = self.fake_get_fc_hbas
        fake_fc_hbas_info.side_effect = self.fake_get_fc_hbas_info

        name = 'volume-00000001'
        vol = {'id': 1, 'name': name}
        location = '10.0.2.15:3260'
        wwn = '1234567890123456'
        connection_info = self.fibrechan_connection(vol, location, wwn)
        conn_data = self.connector._add_targets_to_connection_properties(
            connection_info['data']
        )
        volume_paths = self.connector.get_volume_paths(conn_data)

        expected = ['/dev/disk/by-path/pci-0000:05:00.2'
                    '-fc-0x1234567890123456-lun-1']
        self.assertEqual(expected, volume_paths)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'remove_scsi_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
    def test_connect_volume(self, check_valid_device_mock,
                            get_device_info_mock,
                            get_scsi_wwn_mock,
                            remove_device_mock,
                            get_fc_hbas_info_mock,
                            get_fc_hbas_mock,
                            realpath_mock,
                            exists_mock,
                            wait_for_rw_mock):
        check_valid_device_mock.return_value = True
        get_fc_hbas_mock.side_effect = self.fake_get_fc_hbas
        get_fc_hbas_info_mock.side_effect = self.fake_get_fc_hbas_info

        wwn = '1234567890'
        multipath_devname = '/dev/md-1'
        devices = {"device": multipath_devname,
                   "id": wwn,
                   "devices": [{'device': '/dev/sdb',
                                'address': '1:0:0:1',
                                'host': 1, 'channel': 0,
                                'id': 0, 'lun': 1}]}
        get_device_info_mock.return_value = devices['devices'][0]
        get_scsi_wwn_mock.return_value = wwn

        location = '10.0.2.15:3260'
        name = 'volume-00000001'
        vol = {'id': 1, 'name': name}
        # Should work for string, unicode, and list
        wwns_luns = [
            ('1234567890123456', 1),
            (six.text_type('1234567890123456'), 1),
            (['1234567890123456', '1234567890123457'], 1),
            (['1234567890123456', '1234567890123457'], 1),
        ]
        for wwn, lun in wwns_luns:
            connection_info = self.fibrechan_connection(vol, location,
                                                        wwn, lun)
            dev_info = self.connector.connect_volume(connection_info['data'])
            exp_wwn = wwn[0] if isinstance(wwn, list) else wwn
            dev_str = ('/dev/disk/by-path/pci-0000:05:00.2-fc-0x%s-lun-1' %
                       exp_wwn)
            self.assertEqual(dev_info['type'], 'block')
            self.assertEqual(dev_info['path'], dev_str)
            self.assertNotIn('multipath_id', dev_info)
            self.assertNotIn('devices', dev_info)

            self.connector.disconnect_volume(connection_info['data'], dev_info)
            expected_commands = []
            self.assertEqual(expected_commands, self.cmds)

        # Should not work for anything other than string, unicode, and list
        connection_info = self.fibrechan_connection(vol, location, 123)
        self.assertRaises(exception.VolumePathsNotFound,
                          self.connector.connect_volume,
                          connection_info['data'])

        get_fc_hbas_mock.side_effect = [[]]
        get_fc_hbas_info_mock.side_effect = [[]]
        self.assertRaises(exception.VolumePathsNotFound,
                          self.connector.connect_volume,
                          connection_info['data'])

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device_path')
    def _test_connect_volume_multipath(self, get_device_info_mock,
                                       get_scsi_wwn_mock,
                                       get_fc_hbas_info_mock,
                                       get_fc_hbas_mock,
                                       realpath_mock,
                                       exists_mock,
                                       wait_for_rw_mock,
                                       find_mp_dev_mock,
                                       access_mode,
                                       should_wait_for_rw,
                                       find_mp_device_path_mock):
        self.connector.use_multipath = True
        get_fc_hbas_mock.side_effect = self.fake_get_fc_hbas
        get_fc_hbas_info_mock.side_effect = self.fake_get_fc_hbas_info

        wwn = '1234567890'
        multipath_devname = '/dev/md-1'
        devices = {"device": multipath_devname,
                   "id": wwn,
                   "devices": [{'device': '/dev/sdb',
                                'address': '1:0:0:1',
                                'host': 1, 'channel': 0,
                                'id': 0, 'lun': 1},
                               {'device': '/dev/sdc',
                                'address': '1:0:0:2',
                                'host': 1, 'channel': 0,
                                'id': 0, 'lun': 1}]}
        get_device_info_mock.side_effect = devices['devices']
        get_scsi_wwn_mock.return_value = wwn

        location = '10.0.2.15:3260'
        name = 'volume-00000001'
        vol = {'id': 1, 'name': name}
        initiator_wwn = ['1234567890123456', '1234567890123457']

        find_mp_device_path_mock.return_value = '/dev/mapper/mpatha'
        find_mp_dev_mock.return_value = {"device": "dm-3",
                                         "id": wwn,
                                         "name": "mpatha"}

        connection_info = self.fibrechan_connection(vol, location,
                                                    initiator_wwn)
        connection_info['data']['access_mode'] = access_mode

        self.connector.connect_volume(connection_info['data'])

        self.assertEqual(should_wait_for_rw, wait_for_rw_mock.called)

        self.connector.disconnect_volume(connection_info['data'],
                                         devices['devices'][0])
        expected_commands = [
            'multipath -f ' + find_mp_device_path_mock.return_value,
            'blockdev --flushbufs /dev/sdb',
            'tee -a /sys/block/sdb/device/delete',
            'blockdev --flushbufs /dev/sdc',
            'tee -a /sys/block/sdc/device/delete',
        ]
        self.assertEqual(expected_commands, self.cmds)
        return connection_info

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
    def test_connect_volume_multipath_rw(self, check_valid_device_mock,
                                         get_device_info_mock,
                                         get_scsi_wwn_mock,
                                         get_fc_hbas_info_mock,
                                         get_fc_hbas_mock,
                                         realpath_mock,
                                         exists_mock,
                                         wait_for_rw_mock,
                                         find_mp_dev_mock):

        check_valid_device_mock.return_value = True
        self._test_connect_volume_multipath(get_device_info_mock,
                                            get_scsi_wwn_mock,
                                            get_fc_hbas_info_mock,
                                            get_fc_hbas_mock,
                                            realpath_mock,
                                            exists_mock,
                                            wait_for_rw_mock,
                                            find_mp_dev_mock,
                                            'rw',
                                            True)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
    def test_connect_volume_multipath_no_access_mode(self,
                                                     check_valid_device_mock,
                                                     get_device_info_mock,
                                                     get_scsi_wwn_mock,
                                                     get_fc_hbas_info_mock,
                                                     get_fc_hbas_mock,
                                                     realpath_mock,
                                                     exists_mock,
                                                     wait_for_rw_mock,
                                                     find_mp_dev_mock):

        check_valid_device_mock.return_value = True
        self._test_connect_volume_multipath(get_device_info_mock,
                                            get_scsi_wwn_mock,
                                            get_fc_hbas_info_mock,
                                            get_fc_hbas_mock,
                                            realpath_mock,
                                            exists_mock,
                                            wait_for_rw_mock,
                                            find_mp_dev_mock,
                                            None,
                                            True)

    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
    def test_connect_volume_multipath_ro(self, check_valid_device_mock,
                                         get_device_info_mock,
                                         get_scsi_wwn_mock,
                                         get_fc_hbas_info_mock,
                                         get_fc_hbas_mock,
                                         realpath_mock,
                                         exists_mock,
                                         wait_for_rw_mock,
                                         find_mp_dev_mock):

        check_valid_device_mock.return_value = True
        self._test_connect_volume_multipath(get_device_info_mock,
                                            get_scsi_wwn_mock,
                                            get_fc_hbas_info_mock,
                                            get_fc_hbas_mock,
                                            realpath_mock,
                                            exists_mock,
                                            wait_for_rw_mock,
                                            find_mp_dev_mock,
                                            'ro',
                                            False)

    @mock.patch.object(base.BaseLinuxConnector, '_discover_mpath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
    def test_connect_volume_multipath_not_found(self,
                                                check_valid_device_mock,
                                                get_device_info_mock,
                                                get_scsi_wwn_mock,
                                                get_fc_hbas_info_mock,
                                                get_fc_hbas_mock,
                                                realpath_mock,
                                                exists_mock,
                                                wait_for_rw_mock,
                                                find_mp_dev_mock,
                                                discover_mp_dev_mock):
        check_valid_device_mock.return_value = True
        discover_mp_dev_mock.return_value = ("/dev/disk/by-path/something",
                                             None)

        connection_info = self._test_connect_volume_multipath(
            get_device_info_mock, get_scsi_wwn_mock, get_fc_hbas_info_mock,
            get_fc_hbas_mock, realpath_mock, exists_mock, wait_for_rw_mock,
            find_mp_dev_mock, 'rw', False)

        self.assertNotIn('multipathd_id', connection_info['data'])

    @mock.patch.object(fibre_channel.FibreChannelConnector, 'get_volume_paths')
    def test_extend_volume_no_path(self, mock_volume_paths):
        mock_volume_paths.return_value = []
        volume = {'id': 'fake_uuid'}
        wwn = '1234567890123456'
        connection_info = self.fibrechan_connection(volume,
                                                    "10.0.2.15:3260",
                                                    wwn)

        self.assertRaises(exception.VolumePathsNotFound,
                          self.connector.extend_volume,
                          connection_info['data'])

    @mock.patch.object(linuxscsi.LinuxSCSI, 'extend_volume')
    @mock.patch.object(fibre_channel.FibreChannelConnector, 'get_volume_paths')
    def test_extend_volume(self, mock_volume_paths, mock_scsi_extend):
        fake_new_size = 1024
        mock_volume_paths.return_value = ['/dev/vdx']
        mock_scsi_extend.return_value = fake_new_size
        volume = {'id': 'fake_uuid'}
        wwn = '1234567890123456'
        connection_info = self.fibrechan_connection(volume,
                                                    "10.0.2.15:3260",
                                                    wwn)
        new_size = self.connector.extend_volume(connection_info['data'])
        self.assertEqual(fake_new_size, new_size)

    @mock.patch.object(os.path, 'isdir')
    def test_get_all_available_volumes_path_not_dir(self, mock_isdir):
        mock_isdir.return_value = False
        expected = []
        actual = self.connector.get_all_available_volumes()
        self.assertItemsEqual(expected, actual)

    @mock.patch('eventlet.greenthread.sleep', mock.Mock())
    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
    def test_connect_volume_device_not_valid(self, check_valid_device_mock,
                                             get_device_info_mock,
                                             get_scsi_wwn_mock,
                                             get_fc_hbas_info_mock,
                                             get_fc_hbas_mock,
                                             realpath_mock,
                                             exists_mock,
                                             wait_for_rw_mock,
                                             find_mp_dev_mock):

        check_valid_device_mock.return_value = False
        self.assertRaises(exception.NoFibreChannelVolumeDeviceFound,
                          self._test_connect_volume_multipath,
                          get_device_info_mock,
                          get_scsi_wwn_mock,
                          get_fc_hbas_info_mock,
                          get_fc_hbas_mock,
                          realpath_mock,
                          exists_mock,
                          wait_for_rw_mock,
                          find_mp_dev_mock,
                          'rw',
                          True)

    @ddt.data(
        {
            "target_info": {
                "target_lun": 1,
                "target_wwn": '1234567890123456',
            },
            "expected_targets": [
                ('1234567890123456', 1)
            ]
        },
        {
            "target_info": {
                "target_lun": 1,
                "target_wwn": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 1),
            ]
        },
        {
            "target_info": {
                "target_luns": [1, 1],
                "target_wwn": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 1),
            ]
        },
        {
            "target_info": {
                "target_luns": [1, 2],
                "target_wwn": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 2),
            ]
        },
        {
            "target_info": {
                "target_luns": [1, 1],
                "target_wwns": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 1),
            ]
        },
        {
            "target_info": {
                "target_lun": 7,
                "target_luns": [1, 1],
                "target_wwn": 'foo',
                "target_wwns": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 1),
            ]
        },
        # Add the zone map in now
        {
            "target_info": {
                "target_lun": 1,
                "target_wwn": '1234567890123456',
            },
            "expected_targets": [
                ('1234567890123456', 1)
            ],
            "itmap": {
                '0004567890123456': ['1234567890123456']
            },
            "expected_map": {
                '0004567890123456': [('1234567890123456', 1)]
            }
        },
        {
            "target_info": {
                "target_lun": 1,
                "target_wwn": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 1),
            ],
            "itmap": {
                '0004567890123456': ['1234567890123456',
                                     '1234567890123457']
            },
            "expected_map": {
                '0004567890123456': [('1234567890123456', 1),
                                     ('1234567890123457', 1)]
            }
        },
        {
            "target_info": {
                "target_luns": [1, 2],
                "target_wwn": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 2),
            ],
            "itmap": {
                '0004567890123456': ['1234567890123456'],
                '1004567890123456': ['1234567890123457'],
            },
            "expected_map": {
                '0004567890123456': [('1234567890123456', 1)],
                '1004567890123456': [('1234567890123457', 2)],
            }
        },
        {
            "target_info": {
                "target_luns": [1, 2],
                "target_wwn": ['1234567890123456', '1234567890123457'],
            },
            "expected_targets": [
                ('1234567890123456', 1),
                ('1234567890123457', 2),
            ],
            "itmap": {
                '0004567890123456': ['1234567890123456',
                                     '1234567890123457']
            },
            "expected_map": {
                '0004567890123456': [('1234567890123456', 1),
                                     ('1234567890123457', 2)]
            }
        },

    )
    @ddt.unpack
    def test__add_targets_to_connection_properties(self, target_info,
                                                   expected_targets,
                                                   itmap=None,
                                                   expected_map=None):
        volume = {'id': 'fake_uuid'}
        wwn = '1234567890123456'
        conn = self.fibrechan_connection(volume, "10.0.2.15:3260", wwn)
        conn['data'].update(target_info)

        if itmap:
            conn['data']['initiator_target_map'] = itmap

        connection_info = self.connector._add_targets_to_connection_properties(
            conn['data'])
        self.assertIn('targets', connection_info)
        self.assertEqual(expected_targets, connection_info['targets'])

        if itmap:
            self.assertIn('initiator_target_lun_map', connection_info)
            self.assertEqual(expected_map,
                             connection_info['initiator_target_lun_map'])



# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_brick.initiator.connectors import local
from os_brick.tests.initiator import test_connector


class LocalConnectorTestCase(test_connector.ConnectorTestCase):

    def setUp(self):
        super(LocalConnectorTestCase, self).setUp()
        self.connection_properties = {'name': 'foo',
                                      'device_path': '/tmp/bar'}
        self.connector = local.LocalConnector(None)

    def test_get_connector_properties(self):
        props = local.LocalConnector.get_connector_properties(
            'sudo', multipath=True, enforce_multipath=True)

        expected_props = {}
        self.assertEqual(expected_props, props)

    def test_get_search_path(self):
        actual = self.connector.get_search_path()
        self.assertIsNone(actual)

    def test_get_volume_paths(self):
        expected = [self.connection_properties['device_path']]
        actual = self.connector.get_volume_paths(
            self.connection_properties)
        self.assertEqual(expected, actual)

    def test_connect_volume(self):
        cprops = self.connection_properties
        dev_info = self.connector.connect_volume(cprops)
        self.assertEqual(dev_info['type'], 'local')
        self.assertEqual(dev_info['path'], cprops['device_path'])

    def test_connect_volume_with_invalid_connection_data(self):
        cprops = {}
        self.assertRaises(ValueError,
                          self.connector.connect_volume, cprops)

    def test_extend_volume(self):
        self.assertRaises(NotImplementedError,
                          self.connector.extend_volume,
                          self.connection_properties)
